from Page import Page


class IncludeURLTest(Page):
    """Test includeURL redirection.

    This test will test out the callMethodOfServlet and includeURL
    redirection. The forward() method works similar to these but is
    not tested here. The following operations are performed:

    The form fields are displayed, as seen by this servlet.

    The request environment of this servlet is displayed.

    The writeStatus method on the servlet IncludeURLTest2 which is
    found in the Dir subdirectory under Testing is called.

    The title method is called on several other servlets to demonstrate
    calling methods on servlets in different areas relative to here.

    Finally, the top level page of this context is included with includeURL.

    A 'Test Complete' message is displayed at the very end.
    """

    def writeBody(self):
        self.writeln('<body style="margin:6pt;font-family:sans-serif">')
        fields = self.request().fields()
        name = self.__class__.__name__
        self.writeln(f'<h2>{name}</h2>')
        module = self.__module__
        self.writeln(f'<h3>class = <code>{name}</code>,'
                     f' module= <code>{module}</code></h3>')
        doc = self.__class__.__doc__.replace('\n\n', '</p><p>')
        self.writeln(f'<p>{doc}</p>')
        self.writeln('<h4>Number of fields in the request.fields():'
                     f' {len(fields)}</h4>')
        self.writeln('<ul>')
        enc = self.htmlEncode
        for k, v in fields.items():
            self.writeln(f'<li>{enc(k)} = {enc(v)}</li>')
        self.writeln('</ul>')
        self.writeStatus()
        self.cmos(
            '/Testing/Dir/IncludeURLTest2', 'writeStatus',
            "Expect to see the status written by IncludeURLTest2"
            " which is the same format as the above status,"
            " only relative to /Testing/Dir.")
        self.cmos(
            'Dir/IncludeURLTest2', 'serverSidePath',
            "This returns the serverSide Path of the"
            " Dir/IncludeURLTest2 servlet. Notice that there is"
            " no leading '/' which means this test is relative to"
            " the current directory.")
        self.cmos(
            '/Testing/', 'name',
            "This returns the name of the module at the top of"
            " the Testing context which is 'Main'.")
        self.cmos(
            '/Testing/Main', 'serverSidePath',
            "This returns the serverSidePath of the servlet"
            " accessed at the top of this context.")
        self.cmos(
            'Main', 'serverSidePath',
            "This returns the serverSidePath of the servlet"
            " accessed 'Main' and should be the same as the"
            " servlet accessed through the Testing context.")
        self.writeln('<h4>Including Dir/IncludeURLTest2:</h4>')
        self.write('<div style="margin-left:2em">')
        self.includeURL('Dir/IncludeURLTest2')
        self.write('</div>')
        self.writeln('<h4>Including the Main servlet of the'
                     f' {self.request().contextName()} context:</h4>')
        self.write('<div style="margin-left:2em">')
        self.includeURL('Main')
        self.write('</div>')
        self.writeln(f'<h4>{name} complete.</h4>')
        self.writeln('</body>')

    def writeStatus(self):
        name = self.__class__.__name__
        self.writeln(f'<h4>Request Status of <code>{name}</code>:</h4>')
        w = self.wkv
        req = self.request()
        env = req._environ
        self.writeln("<pre>")
        w('serverSidePath()', req.serverSidePath())
        w('scriptName()', req.scriptName())
        w('servletPath()', req.servletPath())
        w('contextName()', req.contextName())
        w('serverSideContextPath()', req.serverSideContextPath())
        w('extraURLPath()', req.extraURLPath())
        w('urlPath()', req.urlPath())
        w('previousURLPaths()', ', '.join(req.previousURLPaths()))
        w('Environment', '')
        w('REQUEST_URI', env.get('REQUEST_URI', ''))
        w('PATH_INFO', env.get('PATH_INFO', ''))
        w('SCRIPT_NAME', env.get('SCRIPT_NAME', ''))
        w('SCRIPT_FILENAME', env.get('SCRIPT_FILENAME', ''))
        self.writeln('</pre>')

    def wkv(self, k, v):
        self.writeln(f'{k}: {self.htmlEncode(v)}')

    def cmos(self, url, method, desc):
        self.writeln(
            '<p>Calling'
            f' <code>callMethodOfServlet(t, {url!r}, {method!r})</code>:</p>'
            f'<p>{desc}</p>')
        self.write('<div style="margin-left:2em">')
        ret = self.callMethodOfServlet(url, method)
        self.write('</div>')
        ret = 'nothing' if ret is None else f'<code>{ret}</code>'
        self.writeln(
            f'<p><code>callMethodOfServlet</code> returned {ret}.</p>')



"""JSON-RPC servlet base class

Written by Jean-Francois Pieronne
"""

from json import dumps, loads
from io import StringIO
from traceback import print_exc

from HTTPContent import HTTPContent


class JSONRPCServlet(HTTPContent):
    """A superclass for Webware servlets using JSON-RPC techniques.

    JSONRPCServlet can be used to make coding JSON-RPC applications easier.

    Subclasses should override the method json_methods() which returns a list
    of method names. These method names refer to Webware Servlet methods that
    are able to be called by an JSON-RPC-enabled web page. This is very similar
    in functionality to Webware's actions.

    Some basic security measures against JavaScript hijacking are taken by
    default which can be deactivated if you're not dealing with sensitive data.
    You can further increase security by adding shared secret mechanisms.
    """

    # Class level variables that can be overridden by servlet instances:
    _debug = False  # set to True if you want to see debugging output
    # The following variables control security precautions concerning
    # a vulnerability known as "JavaScript hijacking".
    _allowGet = False  # set to True if you want to allow GET requests
    _allowEval = False  # set to True to allow direct evaluation of response

    def __init__(self):
        HTTPContent.__init__(self)

    def respondToGet(self, transaction):
        if self._allowGet:
            self.writeError("GET method not allowed")
        HTTPContent.respondToGet(self, transaction)

    def defaultAction(self):
        self.jsonCall()

    def actions(self):
        actions = HTTPContent.actions(self)
        actions.append('jsonCall')
        return actions

    @staticmethod
    def exposedMethods():
        """Return a list or a set of all exposed RPC methods."""
        return []

    def writeError(self, msg):
        self.write(dumps({'id': self._id, 'code': -1, 'error': msg}))

    def writeResult(self, data):
        data = dumps({'id': self._id, 'result': data})
        if not self._allowEval:
            data = ('throw new Error'
                    f'("Direct evaluation not allowed");\n/*{data}*/')
        self.write(data)

    def jsonCall(self):
        """Execute method with arguments on the server side.

        Returns JavaScript function to be executed by the client immediately.
        """
        self.response().setHeader('Content-Type', 'application/json')
        request = self.request()
        data = loads(request.rawInput().read())
        self._id, call, params = data['id'], data['method'], data['params']
        if call == 'system.listMethods':
            self.writeResult(self.exposedMethods())
        elif call in self.exposedMethods():
            try:
                method = getattr(self, call)
            except AttributeError:
                self.writeError(
                    f'{call}, although an approved method, was not found')
            else:
                try:
                    if self._debug:
                        self.log(f"json call {call}(call)")
                    if isinstance(params, list):
                        result = method(*params)
                    elif isinstance(params, dict):
                        result = method(**params)
                    else:
                        result = method()
                    self.writeResult(result)
                except Exception:
                    err = StringIO()
                    print_exc(file=err)
                    e = err.getvalue()
                    self.writeError(f'{call} was called,'
                                    f' but encountered an error: {e}')
                    err.close()
        else:
            self.writeError(f'{call} is not an approved method')



"""The Transaction container."""

import sys
import traceback


class Transaction:
    """The Transaction container.

    A transaction serves as:

      * A container for all objects involved in the transaction.
        The objects include application, request, response, session
        and servlet.

      * A message dissemination point. The messages include awake(),
        respond() and sleep().

    When first created, a transaction has no session. However, it will
    create or retrieve one upon being asked for session().

    The life cycle of a transaction begins and ends with Application's
    dispatchRequest().
    """

    # region Init

    def __init__(self, application, request=None):
        self._application = application
        self._request = request
        self._response = None
        self._session = None
        self._servlet = None
        self._error = None
        self._nested = 0

    def __repr__(self):
        s = []
        for name in sorted(self.__dict__):
            attr = getattr(self, name)
            if isinstance(attr, type):
                s.append(f'{name}={attr!r}')
        s = ' '.join(s)
        return f'<{self.__class__.__name__} {s}>'

    # endregion Init

    # region Access

    def application(self):
        """Get the corresponding application."""
        return self._application

    def request(self):
        """Get the corresponding request."""
        return self._request

    def response(self):
        """Get the corresponding response."""
        return self._response

    def setResponse(self, response):
        """Set the corresponding response."""
        self._response = response

    def hasSession(self):
        """Return true if the transaction has a session."""
        sid = self._request.sessionId()
        return sid and self._application.hasSession(sid)

    def session(self):
        """Return the session for the transaction.

        A new transaction is created if necessary. Therefore, this method
        never returns None. Use hasSession() if you want to find out if
        a session already exists.
        """
        if not self._session:
            self._session = self._application.createSessionForTransaction(self)
            self._session.awake(self)  # give new servlet a chance to set up
        return self._session

    def setSession(self, session):
        """Set the session for the transaction."""
        self._session = session

    def servlet(self):
        """Return the current servlet that is processing.

        Remember that servlets can be nested.
        """
        return self._servlet

    def setServlet(self, servlet):
        """Set the servlet for processing the transaction."""
        self._servlet = servlet
        if servlet and self._request:
            servlet._serverSidePath = self._request.serverSidePath()

    def duration(self):
        """Return the duration, in seconds, of the transaction.

        This is basically the response end time minus the request start time.
        """
        return self._response.endTime() - self._request.time()

    def errorOccurred(self):
        """Check whether a server error occurred."""
        return isinstance(self._error, Exception)

    def error(self):
        """Return Exception instance if there was any."""
        return self._error

    def setError(self, err):
        """Set Exception instance.

        Invoked by the application if an Exception is raised to the
        application level.
        """
        self._error = err

    # endregion Access

    # region Transaction stages

    def awake(self):
        """Send awake() to the session (if there is one) and the servlet.

        Currently, the request and response do not partake in the
        awake()-respond()-sleep() cycle. This could definitely be added
        in the future if any use was demonstrated for it.
        """
        if not self._nested and self._session:
            self._session.awake(self)
        self._servlet.awake(self)
        self._nested += 1

    def respond(self):
        """Respond to the request."""
        if self._session:
            self._session.respond(self)
        self._servlet.respond(self)

    def sleep(self):
        """Send sleep() to the session and the servlet.

        Note that sleep() is sent in reverse order as awake()
        (which is typical for shutdown/cleanup methods).
        """
        self._nested -= 1
        self._servlet.sleep(self)
        if not self._nested and self._session:
            self._session.sleep(self)
            self._application.sessions().storeSession(self._session)

    # endregion Transaction stages

    # region Debugging

    def dump(self, file=None):
        """Dump debugging info to stdout."""
        if file is None:
            file = sys.stdout
        wr = file.write
        wr(f'>> Transaction: {self}\n')
        for attr in dir(self):
            wr(f'{attr}: {getattr(self, attr)}\n')
        wr('\n')

    # endregion Debugging

    # region Die

    def die(self):
        """End transaction.

        This method should be invoked when the entire transaction is finished
        with. Currently, this is invoked by the Application. This method
        removes references to the different objects in the transaction,
        breaking cyclic reference chains and speeding up garbage collection.
        """
        keys = list(self.__dict__)
        for name in keys:
            delattr(self, name)

    # endregion Die

    # region Exception handling

    _exceptionReportAttrNames = (
        'application request response session servlet').split()

    def writeExceptionReport(self, handler):
        """Write extra information to the exception report."""
        handler.writeTitle(self.__class__.__name__)
        handler.writeAttrs(self, self._exceptionReportAttrNames)

        for name in self._exceptionReportAttrNames:
            obj = getattr(self, '_' + name, None)
            if obj:
                try:
                    obj.writeExceptionReport(handler)
                except Exception:
                    handler.writeln(
                        '<p>Uncaught exception while asking'
                        f' <strong>{name}</strong> to write report:</p>')
                    handler.writeln('<pre>')
                    traceback.print_exc(file=handler)
                    handler.writeln('</pre>')

    # endregion Exception handling



"""DateInterval.py

Convert interval strings (in the form of 1w2d, etc) to seconds, and back again.
Is not exactly about months or years (leap years in particular).

Accepts (y)ear, (b)month, (w)eek, (d)ay, (h)our, (m)inute, (s)econd.

Exports only timeEncode and timeDecode functions.
"""

__all__ = ['timeEncode', 'timeDecode']

import re

from operator import itemgetter

second = 1
minute = second*60
hour = minute*60
day = hour*24
week = day*7
month = day*30
year = day*365
timeValues = {
    'y': year,
    'b': month,
    'w': week,
    'd': day,
    'h': hour,
    'm': minute,
    's': second,
}
timeOrdered = sorted(timeValues.items(), key=itemgetter(1), reverse=True)


def timeEncode(seconds):
    """Encode a number of seconds (representing a time interval).

    Encode the number into a form like 2d1h3s.
    """
    s = []
    for char, amount in timeOrdered:
        if seconds >= amount:
            i, seconds = divmod(seconds, amount)
            s.append(f'{i}{char}')
    return ''.join(s)


_timeRE = re.compile(r'[0-9]+[a-zA-Z]')


def timeDecode(s):
    """Decode a number in the format 1h4d3m (1 hour, 3 days, 3 minutes).

    Decode the format into a number of seconds.
    """
    time = 0
    for match in _timeRE.findall(s):
        char = match[-1].lower()
        try:
            time += int(match[:-1]) * timeValues[char]
        except KeyError:
            raise ValueError(f'Invalid unit of time: {char}') from None
    return time



# pylint: disable=invalid-name

from Page import Page
from WebUtils.Funcs import htmlEncode as enc


class Inspector(Page):

    def writeContent(self):
        req = self.request()
        self.write('Path:<br>\n')
        self.write(f'<code>{enc(req.extraURLPath())}</code><p>\n')
        self.write('Variables:<br>\n')
        self.write('<table>')
        for name in sorted(req.fields()):
            self.write(
                f'<tr><td style="text-align:right">{enc(name)}:</td>'
                f'<td>{enc(req.field(name))}</td></tr>\n')
        self.write('</table><p>\n')
        self.write('Server-side path:<br>\n')
        self.write(f'<code>{enc(req.serverSidePath())}</code><p>\n')



"""Tests various functions of Users and Roles by running example code."""

import shutil
import tempfile
import unittest


class SimpleExampleTest(unittest.TestCase):
    """A simple example to illustrate how to use UserKit."""

    def setUp(self):
        """Make a folder for UserManager data."""
        self._userDataDir = tempfile.mkdtemp()

    def tearDown(self):
        """Remove our test folder for UserManager data."""
        shutil.rmtree(self._userDataDir, ignore_errors=True)

    def testUsersNoRoles(self):
        from UserKit.UserManagerToFile import UserManagerToFile

        userManager = UserManagerToFile()
        userManager.setUserDir(self._userDataDir)

        # Create a user, add to 'staff' role
        fooUser = userManager.createUser('foo', 'bar')

        # bad login
        theUser = userManager.loginName('foo', 'badpass')
        self.assertTrue(
            theUser is None, 'loginName() returns null if login failed.')
        self.assertFalse(
            fooUser.isActive(),
            'User should NOT be logged in since password was incorrect.')

        # good login
        theUser = userManager.loginName('foo', 'bar')
        self.assertTrue(theUser.isActive(), 'User should be logged in now')
        self.assertEqual(
            theUser, fooUser,
            'Should be the same user object, since it is the same user "foo"')

        # logout
        theUser.logout()
        self.assertFalse(
            theUser.isActive(), 'User should no longer be active.')
        self.assertEqual(userManager.numActiveUsers(), 0)

    def testUsersAndRoles(self):
        from UserKit.RoleUserManagerToFile import RoleUserManagerToFile
        from UserKit.HierRole import HierRole
        from hashlib import sha1 as sha

        userManager = RoleUserManagerToFile()
        userManager.setUserDir(self._userDataDir)

        # Setup our roles
        customersRole = HierRole('customers', 'Customers of ACME Industries')
        staffRole = HierRole(
            'staff', 'All staff.'
            ' Staff role includes all permissions of Customers role.',
            [customersRole])

        # Create a user, add to 'staff' role
        # Note that I encrypt my passwords here so they don't appear
        # in plaintext in the storage file.
        johnUser = userManager.createUser('john', sha(b'doe').hexdigest())
        johnUser.setRoles([customersRole])

        fooUser = userManager.createUser('foo', sha(b'bar').hexdigest())
        fooUser.setRoles([staffRole])

        # Check user "foo"
        theUser = userManager.loginName('foo', sha(b'bar').hexdigest())
        self.assertTrue(theUser.isActive(), 'User should be logged in now')
        self.assertEqual(
            theUser, fooUser,
            'Should be the same user object, since it is the same user "foo"')
        self.assertTrue(
            theUser.playsRole(staffRole),
            'User "foo" should be a member of the staff role.')
        self.assertTrue(
            theUser.playsRole(customersRole), 'User "foo" should'
            ' also be in customer role, since staff includes customers.')

        # Check user "John"
        theUser = userManager.loginName('john', sha(b'doe').hexdigest())
        self.assertTrue(theUser.isActive(), 'User should be logged in now.')
        self.assertEqual(
            theUser, johnUser, 'Should be the same user object,'
            ' since it is the same user "John".')
        self.assertFalse(
            theUser.playsRole(staffRole),
            'John should not be a member of the staff.')
        self.assertTrue(
            theUser.playsRole(customersRole),
            'John should play customer role.')



"""Test Webware Example context"""

import os
import unittest

from .AppTest import AppTest


class TestExamples(AppTest, unittest.TestCase):

    settings = dict(
        PrintConfigAtStartUp=False
    )

    def setUp(self):
        AppTest.setUp(self)
        self.testApp.reset()

    @staticmethod
    def removeDemoDatabase():
        if os.path.exists('demo.db'):
            os.remove('demo.db')

    def testStartPage(self):
        r = self.testApp.get('/')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.status_int, 200)
        self.assertEqual(r.content_type, 'text/html')
        self.assertTrue(r.text.startswith('<!DOCTYPE html>'))
        self.assertTrue(r.text.rstrip().endswith('</html>'))
        from sys import version
        pythonVersion = version.split(None, 1)[0]
        webwareVersion = self.app.webwareVersionString()
        r.mustcontain(
            '<html lang="en">', '<head>', '</head>',
            '<title>Welcome</title>', '<meta charset="utf-8">',
            '<style>', 'table.NiceTable {', '</style>',
            '<body style="color:black;background-color:white">', '</body>',
            '<div id="Page">', '<div id="CornerTitle">Webware Examples</div>',
            '<div id="Banner">Welcome</div>', '<div id="Sidebar">',
            '<div class="MenuHeading">Examples</div>',
            '<div class="MenuHeading">Other</div>',
            '<div class="MenuHeading">Contexts</div>',
            '<div class="MenuHeading">E-mail</div>', 'webware-discuss',
            '<div class="MenuHeading">Exits</div>',
            '<a href="https://webwareforpython.github.io/w4py3/">Webware</a>',
            '<a href="https://www.python.org/">Python</a>',
            '<div class="MenuHeading">Versions</div>',
            f'>Webware {webwareVersion}<', f'>Python {pythonVersion}<',
            '<div id="Content">',
            f'<h2>Welcome to Webware for Python {webwareVersion}!</h2>',
            'Along the side of this page you will see various links')

    def testWelcome(self):
        r = self.testApp.get('/Welcome')
        self.assertEqual(r.status, '200 OK')
        version = self.app.webwareVersionString()
        r.mustcontain(
            '<title>Welcome</title>',
            f'<h2>Welcome to Webware for Python {version}!</h2>')

    def testMenu(self):
        r = self.testApp.get('/Examples/')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Examples/')
        r.mustcontain(
            '<title>Welcome</title>',
            '<div class="MenuHeading">Examples</div>')
        r = r.click('^Welcome$')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Examples/Welcome')
        r.mustcontain('<title>Welcome</title>')
        r = r.click('^ShowTime$')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Examples/ShowTime')
        r.mustcontain(
            '<title>ShowTime</title>', 'The current time is:')
        r = r.click('^CountVisits$')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Examples/CountVisits')
        r.mustcontain(
            '<title>CountVisits</title>', '<h3>Counting Visits</h3>')
        r = r.click('^Error')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Examples/Error')
        r.mustcontain(
            '<title>Error raising Example</title>', '<h1>Error Test</h1>')

    def testShowTime(self):
        r = self.testApp.get('/ShowTime')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('<title>ShowTime</title>', 'The current time is:')

    def checkTimes(self, r, n):
        s = '' if n == 1 else 's'
        p = ('<p>You\'ve been here <strong style="background-color:yellow">'
             f'&nbsp;{n}&nbsp;</strong> time{s}.</p>')
        self.assertIn(p, r)

    def testCountVisits(self):
        r = self.testApp.get('/CountVisits')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>CountVisits</title>',
            '<h3>Counting Visits</h3>',
            'This page records your visits using a session object.')
        self.checkTimes(r, 1)
        r = r.click('revisit')
        self.assertEqual(r.status, '200 OK')
        self.checkTimes(r, 2)
        r = r.goto('CountVisits')
        self.assertEqual(r.status, '200 OK')
        self.checkTimes(r, 3)
        r = r.goto('/CountVisits')
        self.assertEqual(r.status, '200 OK')
        self.checkTimes(r, 4)
        r = r.goto('/Examples/CountVisits')
        self.assertEqual(r.status, '200 OK')
        self.checkTimes(r, 5)
        r = r.click('expire')
        self.assertEqual(r.status, '302 Found')
        r = r.follow()
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Examples/CountVisits')
        self.checkTimes(r, 1)
        r = r.goto('CountVisits')
        self.assertEqual(r.status, '200 OK')
        self.checkTimes(r, 2)

    def testError(self):
        r = self.testApp.get('/Error')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Error raising Example</title>',
            '<h1>Error Test</h1>',
            '<form action="Error" method="post">',
            "Don't click this button!",
            no=['Traceback', 'raise error', 'that button!'])
        r = r.form.submit(expect_errors=True)
        self.assertEqual(r.status, '500 Server Error')
        r.mustcontain(
            '<title>Error</title>',
            '<h2 class="section">Error</h2>',
            'The site is having technical difficulties with this page.',
            '<h2 class="section">Traceback</h2>',
            'self.runTransactionViaServlet', 'raise error',
            'RuntimeError: You clicked that button!',
            no=['CustomError', 'SystemError'])
        r = r.goto('/Error')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('<title>Error raising Example</title>')
        r.form['error'].select(text='System Error')
        r = r.form.submit(expect_errors=True)
        self.assertEqual(r.status, '500 Server Error')
        r.mustcontain(
            '<title>Error</title>',
            'SystemError: You clicked that button!',
            no=['RuntimeError', 'CustomError'])
        r = r.goto('/Error')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('<title>Error raising Example</title>')
        r.form['error'].select(text='Custom Error')
        r = r.form.submit(expect_errors=True)
        self.assertEqual(r.status, '500 Server Error')
        r.mustcontain(
            '<title>Error</title>',
            'CustomError: You clicked that button!',
            no=['RuntimeError', 'SystemError'])

    def testView(self):
        r = self.testApp.get('/View')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>View</title>',
            '<h2>View the source of a Webware servlet.</h2>',
            'for your viewing pleasure')
        r = r.click('View source')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path_qs, '/Examples/View?filename=View.py')
        r.mustcontain(
            '<title>', 'View.py</title>',
            '<pre>', '.highlight .kn {',
            '<span class="kn">from</span>', '</pre>')
        r = r.goto('/View?filename=Nada')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Error</title>',
            '<h3 style="color:#a00">Error</h3>',
            "<p>The requested file 'Nada'"
            " does not exist in the proper directory.</p>")
        r = r.goto('/View?filename=Error.py')
        r.mustcontain(
            '<title>', 'Error.py</title>',
            '<span class="nc">Error</span>',
            '<span class="n">ExamplePage</span>')

    def testIntrospect(self):
        r = self.testApp.get('/Introspect')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Introspect</title>',
            'shows the values for various Python expressions',
            'list(globals())', 'ExamplePage')

    def testColors(self):
        r = self.testApp.get('/Colors')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Colors</title>', '<h3>Color Table Demo</h3>',
            '<body style="color:black;background-color:#FFFFFF">',
            '<div style="text-align:center;color:black">',
            no='<div style="text-align:center;color:white">')
        r.form['bgColor'] = '#CC3366'
        r = r.form.submit()
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Colors</title>', '<h3>Color Table Demo</h3>',
            '<body style="color:black;background-color:#CC3366">',
            '<div style="text-align:center;color:white">',
            no='<div style="text-align:center;color:black">')

    def testListBox(self):
        r = self.testApp.get('/ListBox')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>ListBox</title>', '<h2>List box example.</h2>',
            '--- empty ---', no='New item')
        r = r.form.submit('_action_new')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            'List box example', 'New item 1',
            no=['empty', 'item 2'])
        r = r.form.submit('_action_new').form.submit('_action_new')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            'List box example', 'New item 1', 'New item 2', 'New item 3',
            no=['empty', 'item 4'])
        r = r.form.submit('_action_delete')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            'List box example', 'must select a row to delete',
            'New item 1', 'New item 2', 'New item 3')
        r.form['items'].select_multiple(texts=['New item 2'])
        r = r.form.submit('_action_delete')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            'List box example', 'New item 1', 'New item 3',
            no=['must select', 'empty', 'item 2'])
        r.form['items'].select_multiple(texts=['New item 1', 'New item 3'])
        r = r.form.submit('_action_delete')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            'List box example', '--- empty ---',
            no=['must select', 'New item', 'item 1', 'item 2', 'item 3'])
        r.mustcontain('name="items" size="10" style="width:250pt;')
        r = r.form.submit('_action_taller')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('name="items" size="11" style="width:250pt;')
        r = r.form.submit('_action_shorter')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('name="items" size="10" style="width:250pt;')
        r = r.form.submit('_action_wider')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('name="items" size="10" style="width:280pt;')
        r = r.form.submit('_action_narrower')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('name="items" size="10" style="width:250pt;')

    def testForward(self):
        r = self.testApp.get('/Forward')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Forward</title>',
            'This is the <em>Forward</em> servlet speaking.',
            'Hello from IncludeMe')

    def testSecureCountVisits(self):
        r = self.testApp.get('/SecureCountVisits')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Log In</title>', 'Please log in', no='failed')
        r.form['username'] = 'alice'
        r.form['password'] = 'bob'
        r = r.form.submit('login')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Log In</title>', 'Please log in',
            'Login failed', 'try again')
        r.form['username'] = 'alice'
        r.form['password'] = 'alice'
        r = r.form.submit('login')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>SecureCountVisits</title>',
            '<h3>Counting Visits on a Secured Page</h3>',
            'This page records your visits using a session object.',
            'Authenticated user is <strong>alice</strong>.',
            no=['failed', 'log in'])
        self.checkTimes(r, 1)
        r = r.click('Revisit this page')
        self.assertEqual(r.status, '200 OK')
        self.checkTimes(r, 2)
        r = r.click('^SecureCountVisits$')
        self.assertEqual(r.status, '200 OK')
        self.checkTimes(r, 3)
        r = r.goto('/SecureCountVisits')
        self.assertEqual(r.status, '200 OK')
        self.checkTimes(r, 4)
        r = r.click('Logout')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Log In</title>', 'Please log in',
            no='Counting Visits')
        r.form['username'] = 'bob'
        r.form['password'] = 'bob'
        r = r.form.submit('login')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>SecureCountVisits</title>',
            'Authenticated user is <strong>bob</strong>.')
        r = r.click('Logout')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('<title>Log In</title>', no='Counting Visits')

    def testFileUpload(self):
        r = self.testApp.get('/FileUpload')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>File Upload Example</title>',
            '<h1>Upload Test</h1>',
            'This servlet shows how to handle uploaded files.',
            '<form action="FileUpload"', 'multipart/form-data', '</form>',
            no="Here's the file")
        from webtest import Upload  # pylint: disable=import-error
        r.form['filename'] = Upload(
            'README.txt', b'Hello from uploaded file!', 'text/plain')
        r = r.form.submit()
        r.mustcontain(
            '<title>File Upload Example</title>',
            '<h1>Upload Test</h1>',
            "Here's the file you submitted:",
            '<th>name</th><td><strong>README.txt</strong></td>',
            '<tr><th>disposition</th><td>form-data</td></tr>',
            "<tr><th>disposition_options</th><td>"
            "{'name': 'filename', 'filename': 'README.txt'}</td></tr>",
            '<tr><th>headers</th><td>Content-Disposition: form-data;'
            ' name="filename"; filename="README.txt"',
            'Content-Type: text/plain',
            '<tr><th>size</th><td>25 bytes</td></tr>',
            '<tr><th style="vertical-align:top">contents</th>',
            '<pre style=', 'Hello from uploaded file!', '</pre>',
            no='<form')

    def testRequestInformation(self):
        r = self.testApp.get('/RequestInformation')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>RequestInformation</title>',
            '<h3>Request Variables</h3>',
            '<p>The following table shows the values'
            ' for three important request variables.</p>',
            '>environ()<',
            '<td>PATH_INFO</td><td>/RequestInformation</td>',
            '<td>QUERY_STRING</td>', '<td>REQUEST_METHOD</td>',
            '<td>paste.testing</td><td>True</td>',
            '<td>wsgi.input</td>', '<td>wsgi.version</td>',
            no=['fields()', '<td>answer</td>',
                '>cookies()<', '<td>TestCookieName</td>'])
        r = r.click('Reload the page to see them.')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<h3>Request Variables</h3>',
            'The following table shows the values',
            '>environ()<',
            '<td>PATH_INFO</td><td>/RequestInformation</td>',
            '>cookies()<',
            '<td>TestCookieName</td><td>CookieValue</td></tr>',
            '<td>TestAnotherCookie</td><td>AnotherCookieValue</td></tr>',
            no=['fields()', '<td>answer</td>', 'Reload the page'])
        r = r.click('You can also add some test fields.')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<h3>Request Variables</h3>',
            'The following table shows the values',
            '>fields()<',
            '<td>answer</td><td>42</td>',
            "<td>list</td><td>['1',<wbr> '2',<wbr> '3']</td>",
            '<td>question</td><td>what</td>',
            '>environ()<',
            '<td>PATH_INFO</td><td>/RequestInformation</td>',
            'We added some cookies for you',
            '>cookies()<',
            '<td>TestCookieName</td><td>CookieValue</td></tr>',
            no=['Reload the page', 'add some test fields'])
        r = r.goto('/RequestInformation',
                   extra_environ=dict(REMOTE_USER='bob'))
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<h3>Request Variables</h3>',
            '<td>PATH_INFO</td><td>/RequestInformation</td>',
            '<td>REMOTE_USER</td><td>bob</td>')

    def testImageDemo(self):
        r = self.testApp.get('/ImageDemo')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'text/html')
        r.mustcontain(
            '<title>ImageDemo</title>',
            '<h2>Webware Image Generation Demo</h2>',
            '<img src="ImageDemo?fmt=.png"',
            'image has just been generated using the Python Imaging Library')
        r = self.testApp.get('/ImageDemo?fmt=.png')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'image/png')
        self.assertTrue(2000 < r.content_length < 2500)
        self.assertTrue(r.body.startswith(b'\x89PNG\r\n\x1a\n'))

    def testDominateDemo(self):
        r = self.testApp.get('/DominateDemo')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>DominateDemo</title>',
            'Dominate is a Python library',
            'to generate HTML programmatically',
            '<h1>Using Webware with Dominate</h1>',
            '<h2>Hello World!</h2>',
            '<td>One</td>', '<td>Two</td>', '<td>Three</td>',
            'This content has been produced',
            '<a href="https://github.com/Knio/dominate">Dominate</a>')

    def testYattagDemo(self):
        r = self.testApp.get('/YattagDemo')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>YattagDemo</title>',
            'Yattag is a Python library',
            'to generate HTML programmatically',
            '<h1>Using Webware with Yattag</h1>',
            '<h2>Demo contact form</h2>',
            no=['like spam', 'add the subject', 'did not enter'])
        r = r.form.submit()
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<h2>Demo contact form</h2>',
            'Your message looks like spam!',
            no=['add the subject', 'subject looks like spam', 'did not enter'])
        r.form['message'] = ''
        r = r.form.submit()
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<h2>Demo contact form</h2>',
            'You did not enter a message!',
            no=['add the subject', 'like spam'])
        r.form['message'] = 'Hello, world!'
        r = r.form.submit()
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            'Congratulations! You have sent the following message:',
            'Just testing', 'Hello, world!',
            no=['contact form', 'did not enter', 'like spam'])
        r = r.click('Try sending another message')
        r.form['subject'] = ''
        r = r.form.submit()
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<h2>Demo contact form</h2>',
            'Please add the subject of your message.',
            'Your message looks like spam!',
            no=['did not enter', 'subject looks like spam'])
        r.form['subject'] = 'Get free money fast!'
        r.form['message'] = 'That subject sounds fishy.'
        r = r.form.submit()
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<h2>Demo contact form</h2>',
            'Your subject looks like spam!',
            no=['did not enter', 'message looks like spam'])

    def testDBUtilsDemo(self):
        self.removeDemoDatabase()
        r = self.testApp.get('/DBUtilsDemo')
        self.app.addShutDownHandler(self.removeDemoDatabase)
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'text/html')
        r.mustcontain(
            '<title>DBUtils Demo</title>',
            '<h2>Welcome to the DBUtils Demo!</h2>',
            'We are using PooledDB with the sqlite3 database module',
            'Play with the demo database', 'Create tables',
            no=['No database', 'error'])
        r = r.form.submit('_action_createTables')
        r.mustcontain(
            'Creating the following table:',
            'The table was successfully created',
            no=['already exists', 'error'])
        r = r.click('Back')
        r.mustcontain('Play with the demo database')
        r = r.form.submit('_action_addSeminar')
        r.mustcontain('Add a seminar entry to the database:')
        r.form['id'] = 'W4PY'
        r.form['title'] = 'Creating Webware apps'
        r.form['places'] = 12
        r = r.form.submit()
        r.mustcontain('"Creating Webware apps" added to seminars.')
        r = r.click('Back')
        r.mustcontain('Play with the demo database')
        r = r.form.submit('_action_addSeminar')
        r.form['id'] = 'JAVA'
        r.form['title'] = 'Creating Java apps'
        r.form['cost'] = 1000
        r = r.form.submit()
        r.mustcontain('"Creating Java apps" added to seminars.')
        r = r.click('Back')
        r.mustcontain('Play with the demo database')
        r = r.form.submit('_action_listSeminars')
        r.mustcontain(
            'List of seminars in the database:',
            '<td>JAVA</td><td>Creating Java apps</td>'
            '<td align="right">1000</td><td align="right">unlimited</td>',
            '<td>W4PY</td><td>Creating Webware apps</td>'
            '<td align="right">free</td><td align="right">12</td>')
        r.form['id'] = ['JAVA', None]
        r = r.form.submit()
        r.mustcontain('Entries deleted: 1', 'W4PY', no=['JAVA'])
        r = r.click('Back')
        ##
        r.mustcontain('Play with the demo database')
        r = r.form.submit('_action_addAttendee')
        r.mustcontain('Add an attendee entry to the database:')
        r.form['name'] = 'John Doe'
        r.form['paid'] = '1'
        r = r.form.submit()
        r.mustcontain('John Doe added to attendees.')
        r = r.click('Back')
        r.mustcontain('Play with the demo database')
        r = r.form.submit('_action_addAttendee')
        r.form['name'] = "John's Bro"
        r.form['paid'] = '0'
        r = r.form.submit()
        r.mustcontain("John's Bro added to attendees.")
        r = r.click('Back')
        r.mustcontain('Play with the demo database')
        r = r.form.submit('_action_listAttendees')
        r.mustcontain(
            'List of attendees in the database:',
            '<td>John Doe</td><td>Creating Webware apps</td>'
            '<td align="center">Yes</td>',
            "<td>John's Bro</td><td>Creating Webware apps</td>"
            '<td align="center">No</td>')
        r.form['id'] = [None, "W4PY:John's Bro"]
        r = r.form.submit()
        r.mustcontain('Entries deleted: 1', 'John Doe', no=["John's Bro"])
        r = r.click('Back')
        r = r.form.submit('_action_listSeminars')
        r.mustcontain(
            'List of seminars in the database:',
            '<td>W4PY</td><td>Creating Webware apps</td>'
            '<td align="right">free</td><td align="right">11</td>')
        r = r.click('Back')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'text/html')
        r.mustcontain(
            '<title>DBUtils Demo</title>',
            '<h2>Welcome to the DBUtils Demo!</h2>')

    def testAjaxSuggest(self):
        r = self.testApp.get('/AjaxSuggest')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'text/html')
        r.mustcontain(
            '<title>AjaxSuggest</title>',
            '<h2>Ajax "Suggest" Example</h2>',
            'This example uses Ajax techniques to make suggestions',
            'JavaScript enabled', 'Start typing',
            no=['just entered', 'try again'])
        r.form['query'] = 'ajax'
        r = r.form.submit()
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'text/html')
        r.mustcontain(
            '<h2>Ajax "Suggest" Example</h2>',
            'You have just entered the word <b class="in_red">"ajax"</b>.',
            'If you like, you can try again:',
            no=['uses Ajax techniques', 'Start typing'])
        r = r.goto('/AjaxSuggest?_action_=ajaxCall&_call_=suggest&_req_=1&_=a')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'application/json')
        self.assertTrue(r.text.startswith("handleSuggestions(['"))
        self.assertTrue(r.text.endswith("']);"))
        r.mustcontain(no=['AjaxSuggest', 'Ajax "Suggest" Example'])

    def testJSONRPCClient(self):
        r = self.testApp.get('/JSONRPCClient')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'text/html')
        r.mustcontain(
            '<title>JSONRPCClient</title>',
            '<h3>JSON-RPC Example</h3>',
            'This example shows how you can call methods')
        r = self.testApp.post_json('/JSONRPCExample', params={
            "id": 1, "method": "reverse", "params": ["Hello, World!"]})
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'application/json')
        lines = [line.rstrip() for line in r.text.splitlines()]
        self.assertEqual(lines, [
            'throw new Error("Direct evaluation not allowed");',
            '/*{"id": 1, "result": "!dlroW ,olleH"}*/'])
        r = self.testApp.post_json('/JSONRPCExample', params={
            "id": 2, "method": "invalid", "params": ["Hello, World!"]})
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content_type, 'application/json')
        self.assertEqual(
            r.text, '{"id": 2, "code": -1,'
            ' "error": "invalid is not an approved method"}')


class TestExamplesWithoutWSGIWrite(TestExamples):

    settings = dict(
        PrintConfigAtStartUp=False,
        WSGIWrite=False
    )


class TestExamplesWithoutCaching(TestExamples):

    settings = dict(
        PrintConfigAtStartUp=False,
        CacheServletClasses=False,
        CacheServletInstances=False,
        ReloadServletClasses=False
    )



"""The basic Role class."""

from MiscUtils.Funcs import positiveId


class Role:
    """Used in conjunction with RoleUser to provide role-based security.

    All roles have a name and a description and respond to playsRole().

    RoleUser also responds to playsRole() and is the more popular entry point
    for programmers. Application code may then do something along the lines of:

    if user.playsRole('admin'):
        self.displayAdminMenuItems()

    See also:
      * class HierRole
      * class RoleUser
    """

    # region Init

    def __init__(self, name, description=None):
        self._name = name
        self._description = description

    # endregion Init

    # region Attributes

    def name(self):
        return self._name

    def setName(self, name):
        self._name = name

    def description(self):
        return self._description

    def setDescription(self, description):
        self._description = description

    # endregion Attributes

    # region As strings

    def __str__(self):
        return str(self._name)

    def __repr__(self):
        return f'<{self.__class__} {self._name!r} at {positiveId(self):x}>'

    # endregion As strings

    # region The big question

    def playsRole(self, role):
        """Return true if the receiving role plays the role passed in.

        For Role, this is simply a test of equality. Subclasses may override
        this method to provide richer semantics (such as hierarchical roles).
        """
        if not isinstance(role, Role):
            raise TypeError(f'{role} is not a Role object')
        return self == role

    # endregion The big question



import unittest

from MiscUtils.DictForArgs import (
    DictForArgsError, dictForArgs, pyDictForArgs, expandDictWithExtras)


class TestDictForArgs(unittest.TestCase):

    def testPositives(self):
        cases = '''\
# Basics
x=1       == {'x': '1'}
x=1 y=2   == {'x': '1', 'y': '2'}

# Strings
x='a'     == {'x': 'a'}
x="a"     == {'x': 'a'}
x='a b'   == {'x': 'a b'}
x="a b"   == {'x': 'a b'}
x='a"'    == {'x': 'a"'}
x="a'"    == {'x': "a'"}
x="'a'"   == {'x': "'a'"}
x='"a"'   == {'x': '"a"'}

# No value
x         == {'x': '1'}
x y       == {'x': '1', 'y': '1'}
x y=2     == {'x': '1', 'y': '2'}
x=2 y     == {'x': '2', 'y': '1'}
'''
        self._testPositive('', {})
        self._testPositive(' ', {})
        cases = cases.splitlines()
        for case in cases:
            case = case.partition('#')[0].strip()
            if case:
                inp, out = case.split('==', 1)
                out = eval(out)
                self._testPositive(inp, out)

    def _testPositive(self, inp, out):
        res = dictForArgs(inp)
        self.assertEqual(
            res, out,
            f'Expecting: {out!r}\nGot: {res!r}\n')

    def testNegatives(self):
        cases = '''\
-
$
!@#$
'x'=5
x=5 'y'=6
'''
        cases = cases.splitlines()
        for case in cases:
            case = case.partition('#')[0].strip()
            if case:
                self._testNegative(case)

    def _testNegative(self, input_):
        try:
            res = dictForArgs(input_)
        except DictForArgsError:
            pass  # expected
        except Exception as err:
            self.fail(
                f'Expecting DictForArgError.\nGot error: {err!r}.\n')
        else:
            self.fail(
                f'Expecting DictForArgError.\nGot result: {res!r}.\n')

    def testPyDictForArgs(self):
        cases = '''\
x=1 == {'x': 1}
x=1; y=2 == {'x': 1, 'y': 2}
x='a' == {'x': 'a'}
x="a"; y="""b""" == {'x': 'a', 'y': 'b'}
x=(1, 2, 3) == {'x': (1, 2, 3)}
x=['a', 'b'] == {'x': ['a', 'b']}
x='a b'.split() == {'x': ['a', 'b']}
x=['a b'.split(), 1]; y={'a': 1} == {'x': [['a', 'b'], 1], 'y': {'a': 1}}
'''.splitlines()
        for case in cases:
            case = case.strip()
            if case:
                inp, out = case.split('==', 1)
                out = eval(out)
                self.assertEqual(pyDictForArgs(inp), out)

    def testExpandDictWithExtras(self):
        d = {'Name': 'foo', 'Extras': 'x=1 y=2'}
        result = expandDictWithExtras(d)
        self.assertEqual(result, {'Name': 'foo', 'x': '1', 'y': '2'})
        d = {'Name': 'foo', 'bar': 'z = 3'}
        result = expandDictWithExtras(d)
        self.assertEqual(result, d)
        result = expandDictWithExtras(d, key='bar', delKey=False)
        self.assertEqual(result, {'Name': 'foo', 'bar': 'z = 3', 'z': '3'})



"""Session store using memory and files."""

import time
import threading

from MiscUtils import NoDefault

import SessionMemoryStore
import SessionFileStore

from SessionStore import SessionStore

debug = False


class SessionDynamicStore(SessionStore):
    """Stores the session in memory and in files.

    To use this Session Store, set SessionStore in Application.config
    to 'Dynamic'. Other variables which can be set in Application.config are:

    'MaxDynamicMemorySessions', which sets the maximum number of sessions
    that can be in memory at one time. Default is 10,000.

    'DynamicSessionTimeout', which sets the default time for a session to stay
    in memory with no activity. Default is 15 minutes. When specifying this in
    Application.config, use minutes.

    One-shot sessions (usually created by crawler bots) aren't moved to
    FileStore on periodical clean-up. They are still saved on SessionStore
    shutdown. This reduces the number of files in the Sessions directory.
    """

    # region Init

    def __init__(self, app):
        """Create both a file and a memory store."""
        SessionStore.__init__(self, app)
        self._fileStore = SessionFileStore.SessionFileStore(app)
        self._memoryStore = SessionMemoryStore.SessionMemoryStore(
            app, restoreFiles=False)  # session files are read on demand

        # moveToFileInterval specifies after what period of time
        # in seconds a session is automatically moved to a file
        self._moveToFileInterval = app.setting(
            'DynamicSessionTimeout', 15) * 60

        # maxDynamicMemorySessions is what the user actually sets
        # in Application.config, the maximum number of in memory sessions
        self._maxDynamicMemorySessions = app.setting(
            'MaxDynamicMemorySessions', 10000)

        # Used to keep track of sweeping the file store
        self._fileSweepCount = 0

        # Create a re-entrant lock for thread synchronization. The lock is used
        # to protect all code that modifies the contents of the file store and
        # all code that moves sessions between the file and memory stores, and
        # is also used to protect code that searches in the file store for a
        # session. Using the lock in this way avoids a bug that used to be in
        # this code, where a session was temporarily neither in the file store
        # nor in the memory store while it was being moved from file to memory.
        self._lock = threading.RLock()

        if debug:
            print("SessionDynamicStore Initialized")

    # endregion Init

    # region Access

    def __len__(self):
        """Return the number of sessions in the store."""
        with self._lock:
            return len(self._memoryStore) + len(self._fileStore)

    def __getitem__(self, key):
        """Get a session item from the store."""
        # First try to grab the session from the memory store without locking,
        # for efficiency. Only if that fails do we acquire the lock and look
        # in the file store.
        try:
            return self._memoryStore[key]
        except KeyError:
            with self._lock:
                if key in self._fileStore:
                    self.moveToMemory(key)
                # let it raise a KeyError otherwise
                return self._memoryStore[key]

    def __setitem__(self, key, value):
        """Set a sessing item, saving it to the memory store for now."""
        value.setDirty(False)
        self._memoryStore[key] = value

    def __delitem__(self, key):
        """Delete a session item from the memory and the file store."""
        if key not in self:
            raise KeyError(key)
        with self._lock:
            try:
                del self._memoryStore[key]
            except KeyError:
                pass
            try:
                del self._fileStore[key]
            except KeyError:
                pass

    def __contains__(self, key):
        """Check whether the session store has a given key."""
        # First try to find the session in the memory store without locking,
        # for efficiency.  Only if that fails do we acquire the lock and
        # look in the file store.
        if key in self._memoryStore:
            return True
        with self._lock:
            return key in self._memoryStore or key in self._fileStore

    def __iter__(self):
        """Return an iterator over the stored session keys."""
        # since we must be consistent, we cannot chain the iterators
        return iter(self.keys())

    def keys(self):
        """Return a list with all keys of all the stored sessions."""
        with self._lock:
            return self._memoryStore.keys() + self._fileStore.keys()

    def clear(self):
        """Clear the session store in memory and remove all session files."""
        with self._lock:
            self._memoryStore.clear()
            self._fileStore.clear()

    def setdefault(self, key, default=None):
        """Return value if key available, else default (also setting it)."""
        with self._lock:
            try:
                return self[key]
            except KeyError:
                self[key] = default
                return default

    def pop(self, key, default=NoDefault):
        """Return value if key available, else default (also remove key)."""
        with self._lock:
            try:
                return self._memoryStore.pop(key)
            except Exception:
                if default is NoDefault:
                    return self._fileStore.pop(key)
                return self._fileStore.pop(key, default)

    def moveToMemory(self, key):
        """Move the value for a session from file to memory."""
        with self._lock:
            if debug:
                print(f">> Moving {key} to Memory")
            self._memoryStore[key] = self._fileStore.pop(key)

    def moveToFile(self, key):
        """Move the value for a session from memory to file."""
        with self._lock:
            if debug:
                print(f">> Moving {key} to File")
            self._fileStore[key] = self._memoryStore.pop(key)

    def setEncoderDecoder(self, encoder, decoder):
        """Set the serializer and deserializer for the store."""
        SessionStore.setEncoderDecoder(self, encoder, decoder)
        self._fileStore.setEncoderDecoder(encoder, decoder)

    # endregion Access

    # region Application support

    def storeSession(self, session):
        """Save potentially changed session in the store."""
        if self._alwaysSave or session.isDirty():
            key = session.identifier()
            with self._lock:
                if key in self:
                    if key in self._memoryStore:
                        if self._memoryStore[key] is not session:
                            self._memoryStore[key] = session
                    else:
                        self._fileStore[key] = session
                else:
                    self[key] = session

    def storeAllSessions(self):
        """Permanently save all sessions in the store."""
        with self._lock:
            for key in self._memoryStore.keys():
                self.moveToFile(key)

    def cleanStaleSessions(self, task=None):
        """Clean stale sessions.

        Called by the Application to tell this store to clean out all sessions
        that have exceeded their lifetime.
        We want to have their native class functions handle it, though.

        Ideally, intervalSweep would be run more often than the
        cleanStaleSessions functions for the actual stores.
        This may need to wait until we get the TaskKit in place, though.

        The problem is the FileStore.cleanStaleSessions() method can take a
        while to run. So here, we only run the file sweep every fourth time.
        """
        if debug:
            print("Session Sweep started")
        try:
            if self._fileSweepCount == 0:
                self._fileStore.cleanStaleSessions(task)
            self._memoryStore.cleanStaleSessions(task)
        except KeyError:
            pass
        if self._fileSweepCount < 4:
            self._fileSweepCount += 1
        else:
            self._fileSweepCount = 0
        # Now move sessions from memory to file as necessary:
        self.intervalSweep()
        # It's OK for a session to be moved from memory to file or vice versa
        # in between the time we get the keys and the time we actually ask
        # for the session's access time. It may take a while for the fileStore
        # sweep to get completed.

    def intervalSweep(self):
        """The session sweeper interval function.

        The interval function moves sessions from memory to file
        and can be run more often than the full cleanStaleSessions function.
        """
        if debug:
            print("Starting interval Sweep at", time.ctime(time.time()))
            print("Memory Sessions:", len(self._memoryStore),
                  "FileSessions:", len(self._fileStore))
            print("maxDynamicMemorySessions =", self._maxDynamicMemorySessions)
            print("moveToFileInterval =", self._moveToFileInterval)

        now = time.time()

        moveToFileTime = now - self._moveToFileInterval
        keys = []
        for key in self._memoryStore.keys():
            try:
                if self._memoryStore[key].lastAccessTime() < moveToFileTime:
                    if self._memoryStore[key].isNew():
                        if debug:
                            print("trashing one-shot session", key)
                    else:
                        keys.append(key)
            except KeyError:
                pass
        for key in keys:
            try:
                self.moveToFile(key)
            except KeyError:
                pass

        if len(self._memoryStore) > self._maxDynamicMemorySessions:
            keys = self.memoryKeysInAccessTimeOrder()
            excess = len(self._memoryStore) - self._maxDynamicMemorySessions
            if debug:
                print(excess, "sessions beyond the limit")
            keys = keys[:excess]
            for key in keys:
                try:
                    self.moveToFile(key)
                except KeyError:
                    pass

        if debug:
            print("Finished interval Sweep at", time.ctime(time.time()))
            print("Memory Sessions:", len(self._memoryStore),
                  "FileSessions:", len(self._fileStore))

    def memoryKeysInAccessTimeOrder(self):
        """Fetch memory store's keys in ascending order of last access time."""
        return [session.identifier() for session in sorted(
            self._memoryStore.values(), key=lambda v: v.lastAccessTime())]

    # endregion Application support



"""HTTPStatusCodes.py

Dictionary of HTTP status codes.
"""

from http import HTTPStatus

__all__ = [
    'HTTPStatusCodeList', 'HTTPStatusCodes', 'htmlTableOfHTTPStatusCodes']

# pylint: disable=not-an-iterable
HTTPStatusCodeList = [
    (code.value, code.name, code.description) for code in HTTPStatus]

HTTPStatusCodeListColumnNames = ('Code', 'Identifier', 'Description')

# The HTTPStatusCodes can be indexed by either their status code number or
# by a textual identifier. The result is a dictionary with keys code,
# identifier, and description.
HTTPStatusCodes = {}

# Construct HTTPStatusCodes dictionary
for code, identifier, description in HTTPStatusCodeList:
    d = dict(code=code, identifier=identifier, description=description,
             # the following tow exist for backward compatibility only:
             htmlMsg=description, asciiMsg=description)
    HTTPStatusCodes[code] = d
    HTTPStatusCodes[identifier] = d


def htmlTableOfHTTPStatusCodes(
        codes=None,
        tableArgs='', rowArgs='style="vertical-align:top"',
        colArgs='', headingArgs=''):
    """Return an HTML table with HTTP status codes.

    Returns an HTML string containing all the status code information
    as provided by this module. It's highly recommended that if you
    pass arguments to this function, that you do so by keyword.
    """
    if codes is None:
        codes = HTTPStatusCodeList
    tableArgs = ' ' + tableArgs.lstrip() if tableArgs else ''
    rowArgs = ' ' + rowArgs.lstrip() if rowArgs else ''
    headingArgs = ' ' + headingArgs.lstrip() if headingArgs else ''
    colArgs = ' ' + colArgs.lstrip() if colArgs else ''
    res = [f'<table{tableArgs}>', '<tr>']
    res.extend(f'<th{headingArgs}>{heading}</th>'
               for heading in HTTPStatusCodeListColumnNames)
    res.append('</tr>')
    for code, identifier, description in codes:
        res.append(
            f'<tr{rowArgs}><td{colArgs}>{code}</td>'
            f'<td{colArgs}>{identifier}</td><td{colArgs}>{description}</td>'
            '</tr>')
    res.append('</table>')
    return '\n'.join(res)


# Old (deprecated) alias
HTMLTableOfHTTPStatusCodes = htmlTableOfHTTPStatusCodes


if __name__ == '__main__':
    print('''<html>
<head>
    <title>HTTP Status Codes</title>
</head>
<body>
{}
</body>
</html>'''.format(htmlTableOfHTTPStatusCodes()))



from HTTPServlet import HTTPServlet


class Forward1(HTTPServlet):

    def respond(self, transaction):
        transaction.application().forward(
            transaction,
            'Forward1Target' + transaction.request().extraURLPath())



from .ExamplePage import ExamplePage


class CountVisits(ExamplePage):
    """Counting visits example."""

    def writeContent(self):
        if self.request().hasField('expire'):
            self.session().expiring()
            self.sendRedirectAndEnd('CountVisits')
        self.writeln('<h3>Counting Visits</h3>')
        if self.request().isSessionExpired():
            self.writeln('<p><em>Your session has expired.</em></p>')
        count = self.session().value('count', 0) + 1
        self.session().setValue('count', count)
        self.writeln(f"""<p>You've been here
<strong style="background-color:yellow">&nbsp;{count:d}&nbsp;</strong>
time{'' if count == 1 else 's'}.</p>
<p>This page records your visits using a session object.
Every time you <a href="javascript:location.reload()">reload</a>
or <a href="CountVisits">revisit</a> this page, the counter will increase.
If you close your browser or force the session to
<a href="CountVisits?expire=1">expire</a>, then your session will end
and you will see the counter go back to 1 on your next visit.</p>
<p>Try hitting <a href="javascript:location.reload()">reload</a> now.</p>""")



"""Test Webware PSP context"""

import unittest

from .AppTest import AppTest


class TestPSPExamples(AppTest, unittest.TestCase):

    settings = dict(
        PrintConfigAtStartUp=False,
        ClearPSPCacheOnStart=True
    )

    def testStartPage(self):
        r = self.testApp.get('/').click('^PSP$')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/PSP/Examples/')
        self.assertEqual(r.content_type, 'text/html')
        self.assertTrue(r.text.startswith('<!DOCTYPE html>'))
        self.assertTrue(r.text.rstrip().endswith('</html>'))
        r.mustcontain(
            '<title>PSP Hello</title>',
            '<div id="CornerTitle">PSP Examples</div>',
            '<h1 style="text-align:center;color:navy">Hello from PSP!</h1>',
            '<img src="psplogo.png" alt="Python Server Pages">',
            'This is <strong>PSP</strong>.', 'Here are some examples.')

    def testIndex(self):
        r = self.testApp.get('/PSP/Examples/index.psp')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>PSP Hello</title>',
            '<div id="CornerTitle">PSP Examples</div>')

    def testHello(self):
        r = self.testApp.get('/PSP/Examples/').click('^Hello$')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/PSP/Examples/Hello')
        r.mustcontain(
            '<title>PSP Hello</title>',
            '<div id="CornerTitle">PSP Examples</div>')

    def testBraces(self):
        r = self.testApp.get('/PSP/Examples/').click('^Braces')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/PSP/Examples/Braces')
        r.mustcontain(
            '<title>Braces Test</title>',
            '<h4 style="text-align:center;color:navy">'
            'Python Server Pages</h4>',
            '<h2 style="text-align:center">Braces Test</h2>',
            'just use <strong>braces</strong>',
            "<li>I'm number 1</li>", "<li>I'm number 5</li>",
            no="number 6")

    def testSamplePage(self):
        r = self.testApp.get('/PSP/Examples/').click('^PSPTests$', index=0)
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/PSP/Examples/PSPTests')
        r.mustcontain(
            '<title>PSP Tests Page</title>',
            '<h1 style="text-align:center;color:navy">Hello from PSP!</h1>',
            'This is the sample/test page',
            '<p>Hello from included file!<p>',
            'This is an HTML file that is dynamically inserted.',
            '<h4>Below is a loop test:</h4>',
            '<ul><li>Outer loop: 1', '<li>Inner loop: 2',
            '<a href="PSPTests-Braces">braces</a>'
            ' can make things easier here.</p>',
            '<th style="text-align:left">Path Info:</th>',
            '<td>/PSP/Examples/PSPTests</td>',
            '<h4>Request Variables:</h4>',
            '<td style="color:red;font-size:small">REQUEST_METHOD&nbsp;</td>',
            '<td style="color:blue;font-size:small">GET</td>',
            '<h4>Comments:</h4>', 'Nothing should be visible here.',
            '<p>I\'m a method. <b style="color:maroon">Howdy!</b></p>',
            '<b style="color:green">It\'s True</b></p>',
            "<p><strong>That's all, folks.</strong></p>",
            no=['test page uses <a href="Braces"><strong>braces</strong></a>',
                'Comment check', 'not even in Python file'])

    def testSamplePageBraces(self):
        r = self.testApp.get('/PSP/Examples/').click('^PSPTests-Braces$')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/PSP/Examples/PSPTests-Braces')
        r.mustcontain(
            '<title>PSP Tests Page - Braces</title>',
            '<h1 style="text-align:center;color:navy">Hello from PSP!</h1>',
            'This is the sample/test page',
            'test page uses <a href="Braces"><strong>braces</strong></a>',
            '<p>Hello from included file!<p>',
            'This is an HTML file that is dynamically inserted.',
            '<h4>Below is a loop test:</h4>',
            '<ul><li>Outer loop: 1', '<li>Inner loop: 2',
            'The use of braces, while not good for normal Python,'
            ' does make things easier here',
            '<th style="text-align:left">Path Info:</th>',
            '<td>/PSP/Examples/PSPTests-Braces</td>',
            '<h4>Request Variables:</h4>',
            '<td style="color:red;font-size:small">REQUEST_METHOD&nbsp;</td>',
            '<td style="color:blue;font-size:small">GET</td>',
            '<h4>Comments:</h4>', 'Nothing should be visible here.',
            '<p>I\'m a method. <b style="color:maroon">Howdy!</b></p>',
            '<b style="color:green">It\'s True</b></p>',
            "<p><strong>That's all, folks.</strong></p>",
            no=['Comment check', 'not even in Python file'])

    def testMyInclude(self):
        r = self.testApp.get('/PSP/Examples/my_include.psp')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('Hello from included file!')

    def testAnotherInclude(self):
        r = self.testApp.get('/PSP/Examples/PSPinclude.psp')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('The current time is:', '<hr>')



from .AdminSecurity import AdminSecurity


class PlugIns(AdminSecurity):

    def writeContent(self):
        wr = self.writeln
        plugIns = self.application().plugIns()
        if plugIns:
            wr('<h4 style="text-align:center">'
               'The following Plug-ins were found:</h4>')
            wr('<table  class="NiceTable"'
               ' style="margin-left:auto;margin-right:auto">')
            wr('<tr class="TopHeading"><th colspan="3">Plug-ins</th></tr>')
            wr('<tr class="SubHeading">'
               '<th>Name</th><th>Version</th><th>Directory</th></tr>')
            for plugIn in plugIns.values():
                name, path = plugIn.name(), plugIn.path()
                ver = plugIn.properties()['versionString']
                wr(f'<tr><td>{name}</td>'
                   f'<td style="text-align:center">{ver}</td>'
                   f'<td>{path}</td></tr>')
            wr('</table>')
        else:
            wr('<h4 style="text-align:center">No Plug-ins found.</h4>')



"""FieldStorage.py

This module defines a subclass of the standard Python cgi.FieldStorage class
with an extra method that will allow a FieldStorage to parse a query string
even in a POST request.
"""

import cgi
import os

from collections import defaultdict
from urllib import parse


class FieldStorage(cgi.FieldStorage):
    """Modified FieldStorage class for POST requests with query strings.

    Parameters in the query string which have not been sent via POST are
    appended to the field list. This is different from the behavior of
    Python versions before 2.6 which completely ignored the query string in
    POST request, but it's also different from the behavior of the later Python
    versions which append values from the query string to values sent via POST
    for parameters with the same name. With other words, our FieldStorage class
    overrides the query string parameters with the parameters sent via POST.

    As recommended by W3C in section B.2.2 of the HTML 4.01 specification,
    we also support use of ';' in place of '&' as separator in query strings.
    """

    def __init__(self, fp=None, headers=None, outerboundary=b'',
                 environ=None, keep_blank_values=False,
                 strict_parsing=False, limit=None,
                 encoding='utf-8', errors='replace', max_num_fields=None):
        if environ is None:
            environ = os.environ
        method = environ.get('REQUEST_METHOD', 'GET').upper()
        qs_on_post = None if method in ('GET', 'HEAD') else environ.get(
            'QUERY_STRING', None)
        if qs_on_post:
            environ['QUERY_STRING'] = ''
        try:
            if headers is None:
                # work around Python issue 27777 in FieldStorage
                content_type = environ.get('CONTENT_TYPE')
                if (content_type and
                        content_type != 'application/x-www-form-urlencoded'
                        and not content_type.startswith('multipart/')):
                    if 'CONTENT_LENGTH' in environ:
                        headers = {'content-type': content_type,
                                   'content-length': '-1'}
            if max_num_fields is None:
                # max_num_fields is only supported since Python 3.6.7 and 3.7.2
                super().__init__(
                    fp, headers=headers, outerboundary=outerboundary,
                    environ=environ, keep_blank_values=keep_blank_values,
                    strict_parsing=strict_parsing, limit=limit,
                    encoding=encoding, errors=errors)
            else:
                super().__init__(
                    fp, headers=headers, outerboundary=outerboundary,
                    environ=environ, keep_blank_values=keep_blank_values,
                    strict_parsing=strict_parsing, limit=limit,
                    encoding=encoding, errors=errors,
                    max_num_fields=max_num_fields)
        finally:
            if qs_on_post:
                environ['QUERY_STRING'] = qs_on_post
        if qs_on_post:
            self.add_qs(qs_on_post)

    def add_qs(self, qs):
        """Add all non-existing parameters from the given query string."""
        values = defaultdict(list)
        for name_values in qs.split('&'):
            for name_value in name_values.split(';'):
                nv = name_value.split('=', 2)
                if len(nv) != 2:
                    if self.strict_parsing:
                        raise ValueError(f'bad query field: {name_value!r}')
                    continue
                name = parse.unquote(nv[0].replace('+', ' '))
                value = parse.unquote(nv[1].replace('+', ' '))
                if len(value) or self.keep_blank_values:
                    values[name].append(value)
        if self.list is None:
            # This makes sure self.keys() are available, even
            # when valid POST data wasn't encountered.
            self.list = []
        for key in values:
            if key not in self:
                # Only append values that aren't already the FieldStorage;
                # this makes POSTed vars override vars on the query string.
                for value in values[key]:
                    self.list.append(cgi.MiniFieldStorage(key, value))



"""PickleRPC.py

PickleRPC provides a Server object for connection to Pickle-RPC servers
for the purpose of making requests and receiving the responses.

    >>> from MiscUtils.PickleRPC import Server
    >>> server = Server('http://localhost:8080/Examples/PickleRPCExample')
    >>> server.multiply(10,20)
    200
    >>> server.add(10,20)
    30

See also: Server, PickleRPCServlet, Examples.PickleRPCExample

UNDER THE HOOD

Requests look like this::

    {
        'version':    1,  # default
        'action':     'call',  # default
        'methodName': 'NAME',
        'args':       (A, B, ...),  # default = (,)
        'keywords':   {'A': A, 'B': B, ...}  # default = {}
    }

Only 'methodName' is required since that is the only key without a
default value.

Responses look like this::

    {
        'timeReceived': N,
        'timeReponded': M,
        'value': V,
        'exception': E,
        'requestError': E,
    }

'timeReceived' is the time the initial request was received.
'timeResponded' is the time at which the response was finished, as
close to transmission as possible. The times are expressed as
number of seconds since the Epoch, e.g., ``time.time()``.

Value is whatever the method happened to return.

Exception may be 'occurred' to indicate that an exception
occurred, the specific exception, such as "KeyError: foo" or the
entire traceback (as a string), at the discretion of the server.
It will always be a non-empty string if it is present.

RequestError is an exception such as "Missing method
in request." (with no traceback) that indicates a problem with the
actual request received by the Pickle-RPC server.

Value, exception and requestError are all exclusive to each other.

SECURITY

Pickle RPC uses the SafeUnpickler class (in this module) to
prevent unpickling of unauthorized classes.  By default, it
doesn't allow _any_ classes to be unpickled.  You can override
`allowedGlobals()` or `findGlobal()` in a subclass as needed to
allow specific class instances to be unpickled.

Note that both `Transport` in this package and `PickleRPCServlet` in
the Webware main package are derived from `SafeUnpickler`.

CREDIT

The implementation of this module was taken directly from Python's
xmlrpclib and then transformed from XML-orientation to Pickle-orientation.

The zlib compression was adapted from code by Skip Montanaro that I found
here: http://manatee.mojam.com/~skip/python/
"""

__version__ = 1  # PickleRPC protocol version (not the pickle protocol used)

import zlib

from pickle import dumps, Unpickler, UnpicklingError
from io import StringIO
from xmlrpc.client import ProtocolError as _PE


class Error(Exception):
    """The abstract exception/error class for all PickleRPC errors."""


class ResponseError(Error):
    """Unhandled exceptions raised when the server was computing a response.

    These will indicate errors such as:
      * exception in the actual target method on the server
      * malformed responses
      * non "200 OK" status code responses
    """


class ProtocolError(ResponseError, _PE):
    """RPC protocol error."""


class RequestError(Error):
    """Errors originally raised by the server caused by malformed requests."""


class InvalidContentTypeError(ResponseError):
    """Invalid content type error."""

    def __init__(self, headers, content):
        ResponseError.__init__(self)
        self.headers = headers
        self.content = content

    def __repr__(self):
        content = self.content
        return (f'{self.__class__.__name__}:'
                ' Content type is not text/x-python-pickled-dict\n'
                f' headers = {self.headers}\ncontent =\n{content}')

    __str__ = __repr__


class SafeUnpickler:
    """Safe unpickler.

    For security reasons, we don't want to allow just anyone to unpickle
    anything.  That can cause arbitrary code to be executed.
    So this SafeUnpickler base class is used to control what can be unpickled.
    By default it doesn't let you unpickle any class instances at all,
    but you can create subclass that overrides allowedGlobals().

    Note that the PickleRPCServlet class in the Webware package is derived from
    this class and uses its load() and loads() methods to do all unpickling.
    """

    def allowedGlobals(self):
        """Allowed class names.

        Must return a list of (moduleName, klassName) tuples for all
        classes that you want to allow to be unpickled.

        Example::

            return [('datetime', 'date')]

        Allows datetime.date instances to be unpickled.
        """
        return []

    def findGlobal(self, module, klass):
        """Find class name."""
        if (module, klass) not in self.allowedGlobals():
            raise UnpicklingError(
                "For security reasons, you can\'t unpickle"
                f' objects from module {module} with type {klass}.')
        g = {}
        exec(f'from {module} import {klass} as theClass', g)
        return g['theClass']

    def load(self, file):
        """Unpickle a file."""
        safeUnpickler = Unpickler(file)
        safeUnpickler.find_global = self.findGlobal
        return safeUnpickler.load()

    def loads(self, s):
        """Unpickle a string."""
        return self.load(StringIO(s))


class Server:
    """uri [,options] -> a logical connection to an XML-RPC server

    uri is the connection point on the server, given as
    scheme://host/target.

    The standard implementation always supports the "http" scheme.
    If SSL socket support is available, it also supports "https".

    If the target part and the slash preceding it are both omitted,
    "/PickleRPC" is assumed.

    See the module doc string for more information.
    """

    def __init__(self, uri, transport=None, verbose=False, binary=True,
                 compressRequest=True, acceptCompressedResponse=True):
        """Establish a "logical" server connection."""
        # get the url
        import urllib.parse
        import urllib.error
        type_, uri = urllib.parse.splittype(uri)
        if type_ not in ('http', 'https'):
            raise IOError('unsupported Pickle-RPC protocol')
        self._host, self._handler = urllib.parse.splithost(uri)
        if not self._handler:
            self._handler = '/PickleRPC'

        if transport is None:
            transport = (SafeTransport if type_ == 'https' else Transport)()
        self._transport = transport

        self._verbose = verbose
        self._binary = binary
        self._compressRequest = compressRequest
        self._acceptCompressedResponse = acceptCompressedResponse

    def _request(self, methodName, args, keywords):
        """Call a method on the remote server."""
        request = {
            'version':  __version__,
            'action': 'call', 'methodName': methodName,
            'args': args, 'keywords':  keywords
        }
        request = dumps(request, -1 if self._binary else 0)
        if self._compressRequest and len(request) > 1000:
            request = zlib.compress(request, 1)
            compressed = True
        else:
            compressed = False

        response = self._transport.request(
            self._host, self._handler, request,
            verbose=self._verbose, binary=self._binary, compressed=compressed,
            acceptCompressedResponse=self._acceptCompressedResponse)

        return response

    def _requestValue(self, methodName, args, keywords):
        d = self._request(methodName, args, keywords)
        if 'value' in d:
            return d['value']
        if 'exception' in d:
            raise ResponseError(d['exception'])
        if 'requestError' in d:
            raise RequestError(d['requestError'])
        raise RequestError(
            'Response does not have a value, exception or requestError.')

    def __repr__(self):
        return f'<{self.__class__.__name__} for {self._host}{self._handler}>'

    __str__ = __repr__

    def __getattr__(self, name):
        """Magic method dispatcher.

        Note: to call a remote object with an non-standard name,
        use result getattr(server, "strange-python-name")(args)
        """
        return _Method(self._requestValue, name)


ServerProxy = Server  # be like xmlrpclib for those who might guess/expect it


class _Method:
    """Some magic to bind a Pickle-RPC method to an RPC server.

    Supports "nested" methods (e.g. examples.getStateName).
    """

    def __init__(self, send, name):
        self._send = send
        self._name = name

    def __getattr__(self, name):
        return _Method(self._send, f'{self._name}.{name}')

    def __call__(self, *args, **keywords):  # note that keywords are supported
        return self._send(self._name, args, keywords)


# pylint: disable=arguments-differ,invalid-name,unused-argument

class Transport(SafeUnpickler):
    """Handle an HTTP transaction to a Pickle-RPC server."""

    # client identifier (may be overridden)
    user_agent = f'PickleRPC/{__version__} (Webware for Python)'

    def request(self, host, handler, request_body,
                verbose=False, binary=False, compressed=False,
                acceptCompressedResponse=False):
        """Issue a Pickle-RPC request."""

        h = self.make_connection(host)
        if verbose:
            h.set_debuglevel(1)

        self.send_request(h, handler, request_body)
        self.send_host(h, host)
        self.send_user_agent(h)
        self.send_content(
            h, request_body, binary, compressed, acceptCompressedResponse)

        response = h.getresponse()
        h.headers, h.file = response.msg, response.fp

        if response.status != 200:
            raise ProtocolError(
                host + handler, response.status, response.reason, h.headers)

        self.verbose = verbose

        if h.headers['content-type'] not in (
                'text/x-python-pickled-dict',
                'application/x-python-binary-pickled-dict'):
            headers = h.headers.headers
            content = h.file.read()
            raise InvalidContentTypeError(headers, content)

        try:
            content_encoding = h.headers['content-encoding']
            if content_encoding and content_encoding == 'x-gzip':
                return self.parse_response_gzip(h.file)
            if content_encoding:
                raise ProtocolError(
                    host + handler, 500,
                    f'Unknown encoding type: {content_encoding}', h.headers)
            return self.parse_response(h.file)
        except KeyError:
            return self.parse_response(h.file)

    def make_connection(self, host, port=None):
        """Create an HTTP connection object from a host descriptor."""
        import http.client
        return http.client.HTTPConnection(host, port)

    def send_request(self, connection, handler, request_body):
        """Send request."""
        connection.putrequest('POST', handler)

    def send_host(self, connection, host):
        """Send host header."""
        connection.putheader('Host', host)

    def send_user_agent(self, connection):
        """Send user-agent header."""
        connection.putheader('User-Agent', self.user_agent)

    def send_content(self, connection, request_body,
                     binary=False, compressed=False,
                     acceptCompressedResponse=False):
        """Send content."""
        connection.putheader(
            'Content-Type',
            'application/x-python-binary-pickled-dict' if binary
            else 'text/x-python-pickled-dict')
        connection.putheader('Content-Length', str(len(request_body)))
        if compressed:
            connection.putheader('Content-Encoding', 'x-gzip')
        if zlib is not None and acceptCompressedResponse:
            connection.putheader('Accept-Encoding', 'gzip')
        connection.endheaders()
        if request_body:
            connection.send(request_body)

    def parse_response(self, f):
        """Read response from input file and parse it."""
        return self.load(f)

    def parse_response_gzip(self, f):
        """Read response from input file, decompress it, and parse it."""
        return self.loads(zlib.decompress(f.read()))


class SafeTransport(Transport):
    """Handle an HTTPS transaction to a Pickle-RPC server."""

    def make_connection(self, host, port=None, key_file=None, cert_file=None):
        """Create an HTTPS connection object from a host descriptor."""
        try:
            from http.client import HTTPSConnection
        except ImportError as e:
            raise NotImplementedError(
                "Your version of http.client doesn't support HTTPS") from e
        return HTTPSConnection(host, port, key_file, cert_file)



"""The HierRole class."""

from .Role import Role


class HierRole(Role):
    """HierRole is a hierarchical role.

    It points to its parent roles. The hierarchy cannot have cycles.
    """

    def __init__(self, name, description=None, superRoles=None):
        Role.__init__(self, name, description)
        if superRoles is None:
            superRoles = []
        for role in superRoles:
            if not isinstance(role, Role):
                raise TypeError(f'{role} is not a Role object')
        self._superRoles = superRoles[:]

    def playsRole(self, role):
        """Check whether the receiving role plays the role that is passed in.

        This implementation provides for the inheritance supported by HierRole.
        """
        return self == role or any(
            superRole.playsRole(role) for superRole in self._superRoles)



from WebUtils.Funcs import htmlForDict

from .AdminSecurity import AdminSecurity


class Config(AdminSecurity):

    def title(self):
        return 'Config'

    def writeContent(self):
        self.writeln(htmlForDict(
            self.application().config(), topHeading='Application'))



"""Session store in memory."""

from MiscUtils import NoDefault

from SessionStore import SessionStore
from SessionFileStore import SessionFileStore


class SessionMemoryStore(SessionStore):
    """Stores the session in memory as a dictionary.

    This is fast and secure when you have one, persistent application instance.
    """

    # region Init

    def __init__(self, app, restoreFiles=None):
        """Initialize the session memory store.

        If restoreFiles is true, and sessions have been saved to file,
        the store will be initialized from these files.
        """
        SessionStore.__init__(self, app)
        self._store = {}
        if restoreFiles is None:
            restoreFiles = self._retain
        if restoreFiles:
            fileStore = SessionFileStore(app)
            for key in fileStore:
                try:
                    self[key] = fileStore[key]
                except Exception:
                    app.handleException()
            fileStore.clear()
        self._restoreFiles = restoreFiles

    # endregion Init

    # region Access

    def __len__(self):
        """Return the number of sessions in the store."""
        return len(self._store)

    def __getitem__(self, key):
        """Get a session item from the store."""
        return self._store[key]

    def __setitem__(self, key, value):
        """Set a session item, saving it to the store."""
        value.setDirty(False)
        self._store[key] = value

    def __delitem__(self, key):
        """Delete a session item from the store."""
        session = self[key]
        if not session.isExpired():
            session.expiring()
        del self._store[key]

    def __contains__(self, key):
        """Check whether the session store has a given key."""
        return key in self._store

    def __iter__(self):
        """Return an iterator over the stored session keys."""
        return iter(self._store)

    def keys(self):
        """Return a list with the keys of all the stored sessions."""
        return list(self._store)

    def clear(self):
        """Clear the session store, removing all of its items."""
        self._store.clear()

    def setdefault(self, key, default=None):
        """Return value if key available, else default (also setting it)."""
        # note that setdefault() is atomic, so no locking is needed
        return self._store.setdefault(key, default)

    def pop(self, key, default=NoDefault):
        """Return value if key available, else default (also remove key)."""
        # note that pop() is atomic, so no locking is needed
        if default is NoDefault:
            return self._store.pop(key)
        return self._store.pop(key, default)

    # endregion Access

    # region Application support

    def storeSession(self, session):
        """Save already potentially changed session in the store."""
        if self._alwaysSave or session.isDirty():
            key = session.identifier()
            if key not in self or self[key] is not session:
                self[key] = session

    def storeAllSessions(self):
        """Permanently save all sessions in the store."""
        if self._restoreFiles:
            fileStore = SessionFileStore(self._app)
            for key, session in list(self.items()):
                fileStore[key] = session

    # endregion Application support



from os import sep

from .ExamplePage import ExamplePage


class View(ExamplePage):
    """View the source of a Webware servlet.

    For each Webware example, you will see a sidebar with various menu items,
    one of which is "View source of <em>example</em>". This link points to the
    View servlet and passes the filename of the current servlet. The View
    servlet then loads that file's source code and displays it in the browser
    for your viewing pleasure.

    Note that if the View servlet isn't passed a filename,
    it prints the View's docstring which you are reading right now.
    """

    def writeContent(self):
        req = self.request()
        if req.hasField('filename'):
            trans = self.transaction()
            filename = req.field('filename')
            if sep in filename:
                self.write(
                    '<h3 style="color:red">Error</h3>'
                    '<p>Cannot request a file'
                    ' outside of this directory {filename!r}</p>')
                return
            filename = self.request().serverSidePath(filename)
            self.request().fields()['filename'] = filename
            trans.application().forward(trans, 'Colorize.py')
        else:
            doc = self.__class__.__doc__.split('\n', 1)
            doc[1] = '</p>\n<p>'.join(doc[1].split('\n\n'))
            self.writeln('<h2>{}</h2>\n<p>{}</p>'.format(*doc))



from Testing.IncludeURLTest import IncludeURLTest


class IncludeURLTest2(IncludeURLTest):
    """This is the second part of the URL test code.

    It gets included into the IncludeURLTest, and calls methods
    on other servlets to verify the references continue to work.
    """

    def writeBody(self):
        self.writeln('<body style="margin:6pt;font-family:sans-serif">')
        name = self.__class__.__name__
        self.writeln(f'<h2>{name}</h2>')
        module = self.__module__
        self.writeln(f'<h3>class = <code>{name}</code>,'
                     f' module= <code>{module}</code></h3>')
        doc = self.__class__.__doc__.replace('\n\n', '</p><p>')
        self.writeln(f'<p>{doc}</p>')
        self.writeStatus()
        self.cmos(
            "/Testing/", "serverSidePath",
            "Expect to see the serverSidePath of the Testing/Main module.")
        self.writeln('</body>')



from .ExamplePage import ExamplePage


class RequestInformation(ExamplePage):
    """Request information demo."""

    def writeContent(self):
        self.writeln('<h3>Request Variables</h3>')
        request = self.request()
        self.writeln(
            '<p>The following table shows the values'
            ' for three important request variables.</p>')
        p = 'We added some cookies for you.'
        if not self.request().hasCookie('TestCookieName'):
            p += (' <a href="RequestInformation">Reload the page'
                  ' to see them.</a>')
        if not request.fields():
            p += (' <a href="RequestInformation?'
                  'answer=42&list=1&list=2&list=3&question=what">'
                  'You can also add some test fields.</a>')
        self.writeln(f'<p>{p}</p>')
        self.writeln('<table style="font-size:small;width:100%">')
        if request.fields():
            self.showDict('fields()', request.fields())
        self.showDict('environ()', request.environ())
        if request.cookies():
            self.showDict('cookies()', request.cookies())
        self.writeln('</table>')
        setCookie = self.response().setCookie
        setCookie('TestCookieName', 'CookieValue')
        setCookie('TestAnotherCookie', 'AnotherCookieValue')
        setCookie('TestExpire1', 'expires in 1 minute', expires='+1m')
        setCookie('TestExpire2', 'expires in 2 minute', expires='+2m')
        setCookie('TestExpire3', 'expires in 3 minute', expires='+3m')

    def showDict(self, name, d):
        self.writeln(
            '<tr style="vertical-align:top">'
            f'<td style="background-color:#ccf" colspan="2">{name}</td>'
            '</tr>')
        for key in sorted(d):
            html = self.htmlEncode(str(d[key])).replace('\n', '<br>').replace(
                ',', ',<wbr>').replace(';', ';<wbr>').replace(':/', ':<wbr>/')
            self.writeln(
                '<tr style="vertical-align:top;background-color:#eef">'
                f'<td>{key}</td><td>{html}</td></tr>')



"""MiscUtils for Webware for Python"""

__all__ = ['AbstractError', 'NoDefault']


class AbstractError(NotImplementedError):
    """Abstract method error.

    This exception is raised by abstract methods in abstract classes.
    It is a special case of NotImplementedError, that indicates that the
    implementation won't ever be provided at that location in the future
    -- instead the subclass should provide it.

    Typical usage:

        from MiscUtils import AbstractError

        class Foo:
            def bar(self):
                raise AbstractError(self.__class__)

    Note that adding the self.__class__ makes the resulting exception
    *much* more useful.
    """


class NoDefault:
    """Singleton for parameters with no default.

    This provides a singleton "thing" which can be used to initialize
    the "default=" arguments for different retrieval methods.

    For example:

        from MiscUtils import NoDefault
        def bar(self, name, default=NoDefault):
            if default is NoDefault:
                return self._bars[name]  # will raise exception for invalid key
            else:
                return self._bars.get(name, default)

    The value None does not suffice for "default=" because it does not
    indicate whether or not a value was passed.

    Consistently using this singleton is valuable due to subclassing
    situations:

        def bar(self, name, default=NoDefault):
            if someCondition:
                return self.specialBar(name)
            else:
                return SuperClass.bar(name, default)

    It's also useful if one method that uses "default=NoDefault" relies
    on another object and method to which it must pass the default.
    (This is similar to the subclassing situation.)
    """


def installInWebware(_application):
    pass



from HTTPServlet import HTTPServlet


class index(HTTPServlet):

    def respond(self, transaction):
        extraPath = transaction.request().extraURLPath()
        path = transaction.request().urlPath()
        if extraPath and path.endswith(extraPath):
            path = path[:-len(extraPath)]
        if not path.endswith('Welcome'):
            path = path.rpartition('/')[0] + '/Welcome' + extraPath
            # redirection via the server:
            transaction.application().forward(transaction, path)
            # redirection via the client:
            # trans.response().sendRedirect(path)



"""HTTP exceptions.

HTTPExceptions are for situations that are predicted by the HTTP spec.
Where the ``200 OK`` response is typical, a ``404 Not Found``
or ``301 Moved Temporarily`` response is not entirely unexpected.

`Application` catches all `HTTPException` exceptions (and subclasses of
HTTPException), and instead of being errors these are translated into
responses. In various places these can also be caught and changed, for instance
an `HTTPAuthenticationRequired` could be turned into a normal login page.
"""

from WebUtils.Funcs import htmlEncode


class HTTPException(Exception):
    """HTTPException template class.

    Subclasses must define these variables (usually as class variables):

    `_code`:
        a tuple of the integer error code, and the short
        description that goes with it (like ``(200, "OK")``)
    `_description`:
        the long-winded description, to be presented
        in the response page. Or you can override description()
        if you want something more context-sensitive.
    """

    def __str__(self):
        return f'{self.code()} {self.codeMessage()}'

    # region Error codes

    _description = 'An error occurred'

    def code(self):
        """The integer code."""
        return self._code[0]  # pylint: disable=no-member

    def codeMessage(self):
        """The message (like ``Not Found``) that goes with the code."""
        return self._code[1]  # pylint: disable=no-member

    # endregion Error codes

    # region HTML Description

    def html(self):
        """The error page.

        The HTML page that should be sent with the error,
        usually a description of the problem.
        """
        page = '''
<html><head><title>{code} {title}</title></head>
<body>
<h1>{htTitle}</h1>
{body}
</body></html>'''
        return page.format(
            htTitle=self.htTitle(), title=self.title(),
            body=self.htBody(), code=self.code())

    def title(self):
        """The title used in the HTML page."""
        return self.codeMessage()

    def htTitle(self):
        """The title, but it may include HTML markup (like italics)."""
        return self.title()

    def htBody(self):
        """The HTML body of the page."""
        body = self.htDescription()
        if self.args:  # pylint: disable=using-constant-test
            # pylint: disable=not-an-iterable
            body += ''.join(
                '<p>{}</p>\n'.format(htmlEncode(str(p)) for p in self.args))
        return body

    def description(self):
        """Error description.

        Possibly a plain text version of the error description,
        though usually just identical to `htDescription`.
        """
        return self._description

    def htDescription(self):
        """HTML error description.

        The HTML description of the error, for presentation
        to the browser user.
        """
        return self.description()

    # endregion HTML Description

    # region Misc

    def headers(self):
        """Get headers.

        Additional headers that should be sent with the
        response, not including the Status header. For instance,
        the redirect exception adds a Location header.
        """
        return {}

    def setTransaction(self, trans):
        """Set transaction.

        When the exception is caught by `Application`, it tells
        the exception what the transaction is. This way you
        can resolve relative paths, or otherwise act in a manner
        sensitive of the context of the error.
        """
        self._transaction = trans

    # endregion Misc


class HTTPMovedPermanently(HTTPException):
    """HTTPException "moved permanently" subclass.

    When a resource is permanently moved. The browser may remember this
    relocation, and later requests may skip requesting the original
    resource altogether.
    """
    _code = 301, 'Moved Permanently'

    def __init__(self, location=None, webwareLocation=None):
        """Set destination.

        HTTPMovedPermanently needs a destination that you it should be
        directed to -- you can pass `location` *or* `webwareLocation` --
        if you pass `webwareLocation` it will be relative to the Webware
        root location (the mount point of the WSGI application).
        """
        self._location = location
        self._webwareLocation = webwareLocation
        super().__init__()

    def location(self):
        """The location that we will be redirecting to."""
        if self._webwareLocation:
            location = self._transaction.request().servletPath()
            if not self._webwareLocation.startswith('/'):
                location += '/'
            location += self._webwareLocation
        else:
            location = self._location
        return location

    def headers(self):
        """We include a Location header."""
        return {'Location': self.location()}

    def description(self):
        return (
            'The resource you are accessing has been moved to'
            ' <a href="{0}">{0}</a>'.format(htmlEncode(self.location())))


class HTTPTemporaryRedirect(HTTPMovedPermanently):
    """HTTPException "temporary redirect" subclass.

    Like HTTPMovedPermanently, except the redirect is only valid for this
    request. Internally identical to HTTPMovedPermanently, except with a
    different response code. Browsers will check the server for each request
    to see where it's redirected to.
    """
    _code = 307, 'Temporary Redirect'


# This is what people mean most often:
HTTPRedirect = HTTPTemporaryRedirect


class HTTPBadRequest(HTTPException):
    """HTTPException "bad request" subclass.

    When the browser sends an invalid request.
    """
    _code = 400, 'Bad Request'


class HTTPAuthenticationRequired(HTTPException):
    """HTTPException "authentication required" subclass.

    HTTPAuthenticationRequired will usually cause the browser to open up an
    HTTP login box, and after getting login information from the user, the
    browser will resubmit the request. However, this should also trigger
    login pages in properly set up environments (though much code will not
    work this way).

    Browsers will usually not send authentication information unless they
    receive this response, even when other pages on the site have given 401
    responses before. So when using this authentication every request will
    usually be doubled, once without authentication, once with.
    """
    _code = 401, 'Authentication Required'
    _description = "You must log in to access this resource"

    def __init__(self, realm=None):
        if not realm:
            realm = 'Password required'
        if '"' in realm:
            raise ValueError('Realm must not contain quotation marks')
        self._realm = realm
        super().__init__()

    def headers(self):
        return {'WWW-Authenticate': f'Basic realm="{self._realm}"'}


# This is for wording mistakes:
HTTPAuthorizationRequired = HTTPAuthenticationRequired


class HTTPSessionExpired(HTTPException):
    """HTTPException "session expired" subclass.

    This is the same as HTTPAuthenticationRequired, but should be used
    in the situation when a session has expired.
    """
    _code = 401, 'Session Expired'
    _description = 'Your login session has expired - please log in again'


class HTTPForbidden(HTTPException):
    """HTTPException "forbidden" subclass.

    When access is not allowed to this resource. If the user is anonymous,
    and must be authenticated, then HTTPAuthenticationRequired is a preferable
    exception. If the user should not be able to get to this resource (at
    least through the path they did), or is authenticated and still doesn't
    have access, or no one is allowed to view this, then HTTPForbidden would
    be the proper response.
    """
    _code = 403, 'Forbidden'
    _description = "You are not authorized to access this resource"


class HTTPNotFound(HTTPException):
    """HTTPException "not found" subclass.

    When the requested resource does not exist. To be more secretive,
    it is okay to return a 404 if access to the resource is not permitted
    (you are not required to use HTTPForbidden, though it makes it more
    clear why access was disallowed).
    """
    _code = 404, 'Not Found'
    _description = 'The resource you were trying to access was not found'

    def html(self):
        trans = self._transaction
        page = trans.application()._error404
        if page:
            uri = trans.request().uri()
            return page.format(htmlEncode(uri))
        return HTTPException.html(self)


class HTTPMethodNotAllowed(HTTPException):
    """HTTPException "method not allowed" subclass.

    When a method (like GET, PROPFIND, POST, etc) is not allowed
    on this resource (usually because it does not make sense, not
    because it is not permitted). Mostly for WebDAV.
    """
    _code = 405, 'Method Not Allowed'
    _description = 'The method is not supported on this resource'


class HTTPRequestTimeout(HTTPException):
    """HTTPException "request timeout" subclass.

    The client did not produce a request within the time that the
    server was prepared to wait. The client may repeat the request
    without modifications at any later time.
    """
    _code = 408, 'Request Timeout'


class HTTPConflict(HTTPException):
    """HTTPException "conflict" subclass.

    When there's a locking conflict on this resource (in response to
    something like a PUT, not for most other conflicts). Mostly for WebDAV.
    """
    _code = 409, 'Conflict'


class HTTPUnsupportedMediaType(HTTPException):
    """HTTPException "unsupported media type" subclass.

    The server is refusing to service the request because the entity
    of the request is in a format not supported by the requested resource
    for the requested method.
    """
    _code = 415, 'Unsupported Media Type'


class HTTPPreconditionFailed(HTTPException):
    """HTTPException "Precondition Failed" subclass.

    During compound, atomic operations, when a precondition for an early
    operation fail, then later operations in will fail with this code.
    Mostly for WebDAV.
    """
    _code = 412, 'Precondition Failed'


class HTTPServerError(HTTPException):
    """HTTPException "Server Error" subclass.

    The server encountered an unexpected condition which prevented it
    from fulfilling the request.
    """
    _code = 500, 'Server Error'


class HTTPNotImplemented(HTTPException):
    """HTTPException "not implemented" subclass.

    When methods (like GET, POST, PUT, PROPFIND, etc) are not
    implemented for this resource.
    """
    _code = 501, "Not Implemented"
    _description = (
        "The method given is not yet implemented by this application")


class HTTPServiceUnavailable(HTTPException):
    """HTTPException "service unavailable" subclass.

    The server is currently unable to handle the request due to a temporary
    overloading or maintenance of the server. The implication is that this
    is a temporary condition which will be alleviated after some delay.
    """
    _code = 503, "Service Unavailable"
    _description = "The server is currently unable to handle the request"


class HTTPInsufficientStorage(HTTPException):
    """HTTPException "insufficient storage" subclass.

    When there is not sufficient storage, usually in response to a PUT when
    there isn't enough disk space. Mostly for WebDAV.
    """
    _code = 507, 'Insufficient Storage'
    _description = (
        'There was not enough storage space on the server'
        ' to complete your request')



from SessionMemoryStore import SessionMemoryStore
from SessionFileStore import SessionFileStore

from .TestSessionMemoryStore import TestSessionMemoryStore


class SessionFileStoreTest(TestSessionMemoryStore):

    _storeClass = SessionFileStore
    _storeIsOrdered = False

    def testMemoryStoreRestoreFiles(self):
        app = self._app
        store = SessionMemoryStore(app)
        self.assertEqual(len(store), 7)
        self.assertTrue('foo-0' in store and 'foo-6' in store)
        store = SessionMemoryStore(app, restoreFiles=False)
        self.assertEqual(len(store), 0)
        self.assertFalse('foo-0' in store or 'foo-6' in store)

    def testFileStoreRestoreFiles(self):
        app = self._app
        store = SessionFileStore(app)
        self.assertEqual(len(store), 7)
        self.assertTrue('foo-0' in store and 'foo-6' in store)
        store = SessionFileStore(app, restoreFiles=False)
        self.assertEqual(len(store), 0)
        self.assertFalse('foo-0' in store or 'foo-6' in store)



"""Unittests for the Role classes."""

import unittest


class BasicRoleTest(unittest.TestCase):
    """Tests for the basic role class."""

    def roleClasses(self):
        """Return a list of all Role classes for testing."""
        from UserKit.Role import Role
        from UserKit.HierRole import HierRole
        return [Role, HierRole]

    def testRoleBasics(self):
        """Invoke testRole() with each class returned by roleClasses."""
        for roleClass in self.roleClasses():
            self.checkRoleClass(roleClass)

    def checkRoleClass(self, roleClass):
        role = roleClass('foo', 'bar')
        self.assertEqual(role.name(), 'foo')
        self.assertEqual(role.description(), 'bar')
        self.assertEqual(str(role), 'foo')

        role.setName('x')
        self.assertEqual(role.name(), 'x')

        role.setDescription('y')
        self.assertEqual(role.description(), 'y')

        self.assertTrue(role.playsRole(role))


class HierRoleTest(unittest.TestCase):
    """Tests for the hierarchical role class."""

    def testHierRole(self):
        from UserKit.HierRole import HierRole as hr
        animal = hr('animal')
        eggLayer = hr('eggLayer', None, [animal])
        furry = hr('furry', None, [animal])
        # pylint: disable=possibly-unused-variable
        snake = hr('snake', None, [eggLayer])
        dog = hr('dog', None, [furry])
        platypus = hr('platypus', None, [eggLayer, furry])
        vegetable = hr('vegetable')

        roles = locals()
        del roles['hr']
        del roles['self']

        # The tests below are one per line.
        # The first word is the role name. The rest of the words
        # are all the roles it plays (besides itself).
        tests = '''\
            eggLayer animal
            furry animal
            snake eggLayer animal
            dog furry animal
            platypus eggLayer furry animal'''

        tests = [test.split() for test in tests.splitlines()]

        # Strip names
        # Can we use a compounded/nested list comprehension for this?
        oldTest = tests
        tests = []
        for test in oldTest:
            test = [name.strip() for name in test]
            tests.append(test)

        # Now let's actually do some testing...
        for test in tests:
            role = roles[test[0]]
            self.assertTrue(role.playsRole(role))

            # Test that the role plays all the roles listed
            for name in test[1:]:
                playsRole = roles[name]
                self.assertTrue(role.playsRole(playsRole))

            # Now test that the role does NOT play
            # any of the other roles not listed
            otherRoles = roles.copy()
            for name in test:
                del otherRoles[name]
            for name in otherRoles:
                self.assertFalse(role.playsRole(roles[name]))



from .ExamplePage import ExamplePage


class Forward(ExamplePage):

    def writeContent(self):
        trans = self.transaction()
        resp = self.response()
        resp.write(
            "<p>This is the <em>Forward</em> servlet speaking. I am now"
            " going to include the output of the <em>IncludeMe</em> servlet"
            " via Application's <code>includeURL()</code> method:</p>")
        trans.application().includeURL(trans, 'IncludeMe')



#!/usr/bin/env python3

"""MakeAppWorkDir.py

INTRODUCTION

This utility builds a directory tree that can be used as the current
working directory of an instance of your Webware application.
By using a separate directory tree like this, your application can run
without needing write access etc. to the Webware directory tree, and
you can also run more than one application server at once using the
same Webware code. This makes it easy to reuse and keep Webware
updated without disturbing your applications.

USAGE

MakeAppWorkDir.py [Options] WorkDir

Options:
  -c, --context-name=...  The name for the pre-installed context.
                          By default, it will be "MyContext".
  -d, --context-dir=...   The directory where the context will be located,
                          so you can place it outside of the WorkDir.
  -l, --library=...       Other dirs to be included in the search path.
                          You may specify this option multiple times.

WorkDir:
  The target working directory to be created.
"""

import argparse
import glob
import os
import shutil

import webware


defaultContext = 'MyContext'

try:
    chown, chmod = shutil.chown, os.chmod
    import stat
except (AttributeError, ImportError):
    chown = chmod = stat = None


class MakeAppWorkDir:
    """Make a new application runtime directory for Webware.

    This class breaks down the steps needed to create a new runtime
    directory for Webware. That includes all the needed
    subdirectories, default configuration files, and startup scripts.
    Each step can be overridden in a derived class if needed.

    Existing files will not be overwritten, but access permissions
    will be changed accordingly in any case.
    """

    def __init__(self, workDir,
                 contextName=defaultContext, contextDir='', libraryDirs=None,
                 user=None, group=None, verbose=True):
        """Initializer for MakeAppWorkDir.

        Pass in at least the target working directory.
        If you pass None for contextName, then the default context
        will be the the Examples directory as usual.
        """
        self._workDir = os.path.abspath(workDir)
        self._contextName = contextName
        self._contextDir = contextDir
        if libraryDirs is None:
            libraryDirs = []
        self._libraryDirs = libraryDirs
        self._user, self._group = user, group
        self._verbose = verbose
        self._webwareDir = webware.__path__[0]

    def buildWorkDir(self):
        """These are all the steps needed to make a new runtime directory.

        You can override the steps taken here with your own methods.
        """
        if os.path.exists(self._workDir):
            self.msg("The target directory already exists.")
            self.msg(
                "Adding everything needed for a Webware runtime directory...")
        else:
            self.msg("Making a new Webware runtime directory...")
        self.msg()
        self.makeDirectories()
        self.copyConfigFiles()
        self.copyOtherFiles()
        self.setLibDirs()
        if self._contextName:
            self.makeDefaultContext()
        self.addGitIgnore()
        self.changeOwner()
        self.printCompleted()

    def makeDirectories(self):
        """Create all the needed directories if they don't already exist."""
        self.msg("Creating the directory tree...")
        standardDirs = ('', 'Cache', 'Configs', 'ErrorMsgs', 'Logs',
                        'Scripts', 'Sessions', 'Static')
        for path in standardDirs:
            path = os.path.join(self._workDir, path)
            if os.path.exists(path):
                self.msg(f"\t{path} already exists.")
            else:
                os.mkdir(path)
                self.msg(f"\t{path}")
        for path in self._libraryDirs:
            path = os.path.join(self._workDir, path)
            if os.path.exists(path):
                self.msg(f"\t{path} already exists.")
            else:
                os.makedirs(path)
                open(os.path.join(path, '__init__.py'), 'w').write('#\n')
                self.msg(f"\t{path} created.")

    def copyConfigFiles(self):
        """Make a copy of the config files in the Configs directory."""
        self.msg("Copying config files...")
        configs = glob.glob(os.path.join(
            self._webwareDir, 'Configs', '*.config'))
        for name in configs:
            newName = os.path.join(
                self._workDir, "Configs", os.path.basename(name))
            if os.path.exists(newName):
                self.msg(f"\t{newName} already exists.")
            else:
                self.msg(f"\t{newName}")
                shutil.copyfile(name, newName)

    def copyOtherFiles(self):
        """Make a copy of any other necessary files in the new work dir."""
        self.msg("Copying other files...")
        wsgiScriptFile = os.path.join('Scripts', 'WSGIScript.py')
        otherFiles = ('error404.html', wsgiScriptFile)
        for name in otherFiles:
            newName = os.path.join(self._workDir, name)
            if os.path.exists(newName):
                self.msg(f"\t{newName} already exists.")
            else:
                oldName = os.path.join(self._webwareDir, name)
                if os.path.exists(oldName):
                    self.msg(f"\t{newName}")
                    shutil.copyfile(oldName, newName)
                else:
                    self.msg(f"\tWarning: Cannot find {oldName!r}.")

    def setLibDirs(self):
        """Set the library directories in the WSGI script."""
        if not self._libraryDirs:
            return
        self.msg("Setting the library directories...")
        wsgiScript = os.path.join(self._workDir, 'Scripts', 'WSGIScript.py')
        if os.path.isfile(wsgiScript):
            with open(wsgiScript) as f:
                script = f.read()
            if 'libDirs = []' not in script:
                self.msg("\tWarning: Unexpected WSGI script")
            else:
                script = script.replace(
                    'libDirs = []', f'libDirs = {self._libraryDirs!r}')
                with open(wsgiScript, 'w') as f:
                    f.write(script)
        else:
            self.msg("\tWarning: Cannot find WSGI script.")

    def makeDefaultContext(self):
        """Make a very simple context for the newbie user to play with."""
        self.msg("Creating default context...")
        contextDir = os.path.join(
            self._workDir,
            self._contextDir or self._contextName)
        if contextDir.startswith(self._workDir):
            configDir = contextDir[len(self._workDir):]
            while configDir[:1] in (os.sep, os.altsep):
                configDir = configDir[1:]
        else:
            configDir = contextDir
        if os.path.exists(contextDir):
            self.msg(f"\t{contextDir} already exists.")
        else:
            self.msg(f"\t{contextDir}")
            os.makedirs(contextDir)
        for name in exampleContext:
            filename = os.path.join(contextDir, name)
            if os.path.exists(filename):
                self.msg(f"\t{filename} already exists.")
            else:
                self.msg(f"\t{filename}")
                open(filename, "w").write(exampleContext[name])
        self.msg("Updating config for default context...")
        filename = os.path.join(
            self._workDir, 'Configs', 'Application.config')
        self.msg(f"\t{filename}")
        content = open(filename).readlines()
        foundContext = 0
        with open(filename, 'w') as output:
            for line in content:
                contextName = self._contextName
                if line.startswith(
                        f"Contexts[{contextName!r}] = {configDir!r}\n"):
                    foundContext += 1
                elif line.startswith("Contexts['default'] = "):
                    output.write(
                        f"Contexts[{contextName!r}] = {configDir!r}\n")
                    output.write(f"Contexts['default'] = {contextName!r}\n")
                    foundContext += 2
                else:
                    output.write(line)
        if foundContext < 2:
            self.msg("\tWarning: Default context could not be set.")

    def addGitIgnore(self):
        self.msg("Creating .gitignore file...")
        existed = False
        ignore = '*~ *.bak *.default *.log *.patch *.pstats *.pyc *.ses *.swp'
        ignore = '\n'.join(ignore.split()) + '\n\n__pycache__/\n'
        ignoreDirs = 'Cache ErrorMsgs Logs Sessions'
        ignore += '\n' + '\n'.join(f'/{d}/' for d in ignoreDirs.split()) + '\n'
        filename = os.path.join(self._workDir, '.gitignore')
        if os.path.exists(filename):
            existed = True
        else:
            with open(filename, 'w') as f:
                f.write(ignore)
        if existed:
            self.msg("\tDid not change existing .gitignore file.")

    def changeOwner(self):
        owner = {}
        if self._user:
            owner['user'] = self._user
        if self._group:
            owner['group'] = self._group
        if not owner:
            return
        if not chown:
            self.msg("\tCannot change ownership on this system.")
            return
        self.msg("Changing the ownership...")
        mode = (
            stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
            stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP)
        try:
            path = self._workDir
            chown(path, **owner)
            chmod(path, os.stat(path).st_mode & mode)
            for dirPath, dirNames, fileNames in os.walk(self._workDir):
                for name in dirNames + fileNames:
                    path = os.path.join(dirPath, name)
                    chown(path, **owner)
                    chmod(path, os.stat(path).st_mode & mode)
        except Exception as e:
            self.msg("\tWarning: The ownership could not be changed.")
            self.msg("\tYou may need superuser privileges to do that.")
            self.msg(f"\tThe error message was: {e}")
            return
        mode = 0
        if self._user:
            mode |= stat.S_IWUSR
        if self._group:
            mode |= stat.S_ISGID | stat.S_IWGRP
        try:
            writeDirs = ('Cache', 'ErrorMsgs', 'Logs', 'Sessions')
            for dirName in writeDirs:
                path = os.path.join(self._workDir, dirName)
                chmod(path, os.stat(path).st_mode | mode)
        except Exception as e:
            self.msg("\tWarning: Write permissions could not be set.")
            self.msg(f"\tThe error message was: {e}")

    def printCompleted(self):
        print("""
Congratulations, you've just created a runtime working directory for Webware.

To start the development server, run this command:

webware serve

If you want to open your web browser after running the server, you can
add the "-b" option. Use the "-h" option to see all other possible options.

In a productive environment, we recommend using Apache and mod_wsgi instead
of the waitress development server, and serving the static assets directly
via the Apache server. The Scripts directory contains a WSGI script that can
also be used with any other WSGI server as application server, and any other
web server as reverse proxy which can also be used to serve static assets.

Have fun!""")

    def msg(self, text=None):
        if self._verbose:
            if text:
                print(text)
            else:
                print()


exampleContext = {  # files copied to example context

    # This is used to create a very simple sample context for the new
    # work dir to give the newbie something easy to play with.

    '__init__.py': r"""
def contextInitialize(application, path):
    # You could put initialization code here to be executed
    # when the context is loaded into Webware.
    pass
""",

    'Main.py': r"""
from Page import Page


class Main(Page):

    def title(self):
        return 'My Sample Context'

    def htBodyArgs(self):
        return ('style="color:#202040;background-color:#e8e8f0;'
                'font-family:Tahoma,Verdana,Arial,Helvetica,sans-serif;'
                'font-size:12pt;line-height:1.5;padding:2em"')

    def writeContent(self):
        self.writeln('<h1>Welcome to Webware for Python!</h1>')
        self.writeln('''
        <p>
        This is a sample context generated for you that has purposely been
        kept very simple to give you something to play with to get yourself
        started. The code that implements this page is located in <b>{}</b>.
        </p>'''.format(self.request().serverSidePath()))
        self.writeln('''
        <p>
        There are more examples and documentation in the Webware distribution,
        which you can get to from here:</p>
        <ul>
        ''')
        servletPath = self.request().servletPath()
        contextName = self.request().contextName()
        for ctx in sorted(self.application().contexts()):
            if ctx in ('default', contextName) or '/' in ctx:
                continue
            self.writeln(f'<li><a href="{servletPath}/{ctx}/">{ctx}</a></li>')
        self.writeln('</ul>')
"""

}  # end of example context files


def make(args):
    workDir = args.working_directory
    contextName = args.context_name
    contextDir = args.context_dir
    if not contextName:
        if contextDir:
            contextName = os.path.basename(contextDir)
        else:
            contextName = defaultContext
    libraryDirs = args.library
    if chown:
        user, group = args.user, args.group
    else:
        user = group = None
    command = MakeAppWorkDir(
        workDir, contextName, contextDir, libraryDirs, user, group)
    command.buildWorkDir()


def addArguments(parser):
    """Add command line arguments to the given parser."""
    parser.add_argument(
        '-c', '--context-name',
        help='the name for the pre-installed context'
        f' (by default, the name will be {defaultContext!r})')
    parser.add_argument(
        '-d', '--context-dir',
        help='the directory where the context will be located'
        ' (so you can place it outside the WORK_DIR)')
    parser.add_argument(
        '-l', '--library', action='append',
        help='other directories to be included in the search path'
        ' (you may specify this option multiple times)')
    if chown:
        parser.add_argument(
            '-u', '--user',
            help='the user that shall own the files'
            ' (name or id, e.g. user running the web server)')
        parser.add_argument(
            '-g', '--group',
            help='the group that shall own the files'
            ' (name or id, e.g. group running the web server)')
    parser.add_argument('working_directory', metavar='WORK_DIR')


def main(args=None):
    """Evaluate the command line arguments and call make()."""
    parser = argparse.ArgumentParser(
        description="Make a Webware application working directory")
    addArguments(parser)
    args = parser.parse_args(args)
    make(args)


if __name__ == '__main__':
    main()



"""Automated tests for the PSP BraceConverter

Contributed 2000-09-04 by Dave Wallace
"""

import unittest
from io import StringIO

from PSP.BraceConverter import BraceConverter
from PSP.ServletWriter import ServletWriter


class DummyWriter(ServletWriter):  # pylint: disable=abstract-method
    """Dummy writer for testing."""

    def __init__(self):  # pylint: disable=super-init-not-called
        self._file = StringIO()
        self._tabCount = 3  # base indentation of our test examples
        self._blockCount = 0
        self._indentSpaces = ServletWriter._spaces
        self._useTabs = False
        self._useBraces = False
        self._indent = '    '
        self._userIndent = ServletWriter._emptyString

    def getOutput(self):
        return self._file.getvalue()

    def close(self):
        self._file.close()


class TestBraceConverter(unittest.TestCase):

    @staticmethod
    def trim(text):
        return '\n'.join(filter(None, map(str.rstrip, text.splitlines())))

    def assertParses(self, psp, expected):
        dummyWriter = DummyWriter()
        braceConverter = BraceConverter()
        for line in psp.splitlines():
            braceConverter.parseLine(line, dummyWriter)
        output = dummyWriter.getOutput()
        dummyWriter.close()
        output, expected = list(map(self.trim, (output, expected)))
        self.assertEqual(
            output, expected,
            f'\n\nOutput:\n{output}\n\nExpected:\n{expected}\n')

    def testSimple(self):
        self.assertParses('''
            if a == b: { return True } else: { return False }
        ''', '''
            if a == b:
                return True
            else:
                return False
        ''')

    def testDict(self):
        self.assertParses('''
            for x in range(10): { q = {
            'test': x
            }
            print(x)
            }
        ''', '''
            for x in range(10):
                q = {
                'test': x
                }
                print(x)
        ''')

    def testNestedDict(self):
        self.assertParses(r'''
            if x: { q = {'test': x}; print(x)} else: { print("\"done\"") #""}{
            x = { 'test1': {'sub2': {'subsub1': 2}} # yee ha
            }
            } print("all done")
        ''', r'''
            if x:
                q = {'test': x}; print(x)
            else:
                print("\"done\"") #""}{
                x = { 'test1': {'sub2': {'subsub1': 2}} # yee ha
                }
            print("all done")
        ''')



"""The PSP parser.

This module handles the actual reading of the characters in the source
PSP file and checking it for valid psp tokens. When it finds one,
it calls ParseEventHandler with the characters it found.

Copyright (c) by Jay Love, 2000 (mailto:jsliv@jslove.org)

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation or portions thereof, including modifications,
that you make.

This software is based in part on work done by the Jakarta group.
"""

from io import StringIO

from .PSPUtils import checkAttributes, PSPParserException


checklist = []


def checker(method):
    """Decorator for adding a method to the checklist."""
    checklist.append(method)
    return method


class PSPParser:
    """The main PSP parser class.

    The PSPParser class does the actual sniffing through the input file
    looking for anything we're interested in. Basically, it starts by
    looking at the code looking for a '<' symbol. It looks at the code by
    working with a PSPReader object, which handles the current location in
    the code. When it finds one, it calls a list of checker methods,
    asking each if it recognizes the characters as its kind of input.
    When the checker methods look at the characters, if they want it,
    they go ahead and gobble it up and set up to create it in the servlet
    when the time comes. When they return, they return true if they accept
    the character, and the PSPReader object cursor is positioned past the
    end of the block that the checker method accepted.
    """

    checklist = checklist  # global list of checker methods

    def __init__(self, ctxt):
        self._reader = ctxt.getReader()
        self._writer = ctxt.getServletWriter()
        self._handler = None
        self.cout = StringIO()  # for dumping HTML that none of the check wants
        self.tmplStart = None  # marks the start of HTML code
        self.tmplStop = None  # marks the end of HTML code
        self.currentFile = self._reader.mark().getFile()

    def setEventHandler(self, handler):
        """Set the handler this parser will use when it finds PSP code."""
        self._handler = handler

    def flushCharData(self, start, stop):
        """Dump everything to the char data handler.

        Dump all the HTML that we've accumulated over to the character data
        handler in the event handler object.
        """
        data = self.cout.getvalue()
        self.cout.close()
        if data:  # make sure there's something there
            self._handler.handleCharData(start, stop, data)
        self.cout = StringIO()

    @checker
    def commentCheck(self, _handler, reader):
        """Comments just get eaten."""
        if reader.matches('<%--'):
            reader.advance(4)
            if reader.skipUntil('--%>') is None:
                raise PSPParserException('Comment not terminated')
            self.flushCharData(self.tmplStart, self.tmplStop)
            return True
        return False

    @checker
    def checkExpression(self, handler, reader):
        """Look for "expressions" and handle them."""
        if not reader.matches('<%='):
            return False
        reader.advance(3)  # eat the opening tag
        reader.peekChar()
        reader.skipSpaces()
        start = reader.mark()
        stop = reader.skipUntil('%>')
        if stop is None:
            raise PSPParserException('Expression not terminated')
        handler.setTemplateInfo(self.tmplStart, self.tmplStop)
        handler.handleExpression(start, stop, None)
        return True

    @checker
    def checkDirective(self, handler, reader):
        """Check for directives; for now we support only page and include."""
        if not reader.matches('<%@'):
            return False
        start = reader.mark()
        reader.advance(3)
        reader.skipSpaces()
        for directive in ('page', 'include', 'taglib'):
            if reader.matches(directive):
                match = directive
                break
        else:
            raise PSPParserException('Invalid directive')
        reader.advance(len(match))
        # parse the directive attr:val pair dictionary
        attrs = reader.parseTagAttributes()
        if match == 'page':
            checkAttributes('Page directive', attrs, ([], {
                'imports', 'extends', 'method',
                'isThreadSafe', 'isInstanceSafe',
                'indentType', 'indentSpaces',
                'gobbleWhitespace', 'formatter'}))
        elif match == 'include':
            checkAttributes('Include directive', attrs, (['file'], []))
        else:
            raise PSPParserException(f'{match} directive not implemented')
        reader.skipSpaces()  # skip to where we expect a close tag
        if reader.matches('%>'):
            reader.advance(2)  # advance past it
        else:
            raise PSPParserException('Directive not terminated')
        stop = reader.mark()
        handler.setTemplateInfo(self.tmplStart, self.tmplStop)
        handler.handleDirective(match, start, stop, attrs)
        return True

    @checker
    def checkEndBlock(self, handler, reader):
        """Check for the end of a block."""
        start = reader.mark()
        if reader.matches('<%'):
            reader.advance(2)
            reader.skipSpaces()
            if reader.matches('end'):
                reader.advance(3)
                reader.skipSpaces()
                if reader.matches('%>'):
                    reader.advance(2)
                    handler.setTemplateInfo(self.tmplStart, self.tmplStop)
                    handler.handleEndBlock()
                    return True
                if reader.matches('$%>'):
                    reader.advance(3)
                    handler.setTemplateInfo(self.tmplStart, self.tmplStop)
                    handler.handleEndBlock()
                    print('INFO: A $ at the end of an end tag does nothing.')
                    return True
        # that wasn't it
        reader.reset(start)
        return False

    @checker
    def checkScript(self, handler, reader):
        """The main thing we're after. Check for embedded scripts."""
        if not reader.matches('<%'):
            return False
        reader.advance(2)
        # don't skip as spaces may be significant; leave this for the generator
        start = reader.mark()
        try:
            stop = reader.skipUntil('%>')
        except EOFError as e:
            raise EOFError(
                "Reached EOF while looking for ending script tag") from e
        if stop is None:
            raise PSPParserException('Script not terminated')
        handler.setTemplateInfo(self.tmplStart, self.tmplStop)
        handler.handleScript(start, stop, None)
        return True

    @checker
    def checkScriptFile(self, handler, reader):
        """Check for file level code.

        Check for Python code that must go to the top of the generated module::

            <psp:file>
                import xyz
                print('hi Mome!')
                def foo(): return 'foo'
            </psp:file>
        """
        if not reader.matches('<psp:file>'):
            return False
        reader.advance(10)
        start = reader.mark()
        try:
            stop = reader.skipUntil('</psp:file>')
            if stop is None:
                raise PSPParserException(
                    'Script not terminated in <psp:file> block')
        except EOFError as e:
            raise EOFError(
                'Reached EOF while looking for'
                ' ending script tag </psp:file>') from e
        handler.setTemplateInfo(self.tmplStart, self.tmplStop)
        handler.handleScriptFile(start, stop, None)
        return True

    @checker
    def checkScriptClass(self, handler, reader):
        """Check for class level code.

        Check for Python code that should go in the class definition::

            <psp:class>
                def foo(self):
                    return self.dosomething()
            </psp:class>
        """
        if not reader.matches('<psp:class>'):
            return False
        reader.advance(11)
        start = reader.mark()
        try:
            stop = reader.skipUntil('</psp:class>')
            if stop is None:
                raise PSPParserException(
                    'Script not terminated in <psp:class> block')
        except EOFError as e:
            raise EOFError(
                'Reached EOF while looking for'
                ' ending script tag </psp:class>') from e
        handler.setTemplateInfo(self.tmplStart, self.tmplStop)
        handler.handleScriptClass(start, stop, None)
        return True

    @checker
    def checkMethod(self, handler, reader):
        """Check for class methods defined in the page.

        We only support one format for these,
        ``<psp:method name="xxx" params="xxx,xxx">``
        Then the function body, then </psp:method>.
        """
        if not reader.matches('<psp:method'):
            return False
        start = reader.mark()
        reader.advance(11)
        attrs = reader.parseTagAttributes()
        checkAttributes('method', attrs, (['name'], ['params']))
        reader.skipSpaces()
        if not reader.matches('>'):
            raise PSPParserException('Expected method declaration close')
        reader.advance(1)
        stop = reader.mark()
        handler.setTemplateInfo(self.tmplStart, self.tmplStop)
        handler.handleMethod(start, stop, attrs)
        start = stop
        # skip past the close marker, return the point before the close marker
        stop = reader.skipUntil('</psp:method>')
        handler.handleMethodEnd(start, stop, attrs)
        return True

    @checker
    def checkInclude(self, handler, reader):
        """Check for inserting another pages output in this spot."""
        if not reader.matches('<psp:include'):
            return False
        reader.advance(12)
        reader.skipSpaces()
        attrs = reader.parseTagAttributes()
        checkAttributes('include', attrs, (['path'], []))
        reader.skipSpaces()
        if not reader.matches('>'):
            raise PSPParserException('Include bodies not implemented')
        reader.advance(1)
        handler.setTemplateInfo(self.tmplStart, self.tmplStop)
        handler.handleInclude(attrs, None)
        return True

    @checker
    def checkInsert(self, handler, reader):
        """Check for straight character dumps.

        No big hurry for this. It's almost the same as the page include
        directive.  This is only a partial implementation of what JSP does.
        JSP can pull it from another server, servlet, JSP page, etc.
        """
        if not reader.matches('<psp:insert'):
            return False
        reader.advance(11)
        reader.skipSpaces()
        attrs = reader.parseTagAttributes()
        checkAttributes('insert', attrs, (['file'], []))
        reader.skipSpaces()
        if not reader.matches('>'):
            raise PSPParserException('Insert bodies not implemented')
        reader.advance(1)
        handler.setTemplateInfo(self.tmplStart, self.tmplStop)
        handler.handleInsert(attrs, None)
        return True

    def parse(self, until=None):
        """Parse the PSP file."""
        reader = self._reader
        handler = self._handler
        noPspElement = False
        while reader.hasMoreInput():
            # This is for XML style blocks, which we're not handling yet:
            if until and reader.matches(until):
                return
            # If the file the reader is working on has changed due to
            # a push or pop, flush any char data from the old file:
            if reader.mark().getFile() != self.currentFile:
                self.flushCharData(self.tmplStart, self.tmplStop)
                self.currentFile = reader.mark().getFile()
                self.tmplStart = reader.mark()
            for checkFunc in self.checklist:
                if checkFunc(self, handler, reader):
                    noPspElement = False
                    break
            else:
                if not noPspElement:
                    self.tmplStart = reader.mark()
                    noPspElement = True
                s = reader.nextContent()  # skip till the next possible tag
                self.tmplStop = reader.mark()  # mark the end of HTML data
                self.cout.write(s)  # write out the raw HTML data
            self.flushCharData(self.tmplStart, self.tmplStop)  # dump the rest



"""ImportManager

Manages imported modules and protects against concurrent imports.

Keeps lists of all imported Python modules and templates as well as other
config files used by Webware for Python. Modules which are not directly
imported can be monitored using hupper. This can be used to detect changes in
source files, templates or config files in order to reload them automatically.
"""


import sys
import warnings

from os.path import getmtime, isfile, join
from importlib.util import module_from_spec, spec_from_file_location
from importlib.machinery import (
    ModuleSpec, SourceFileLoader, SourcelessFileLoader)


class ImportManager:
    """The import manager.

    Keeps track of the Python modules and other system files that have been
    imported and are used by Webware.
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        """Create ImportManager as a singleton."""
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        """Initialize import manager."""
        self._fileList = {}
        self._moduleFiles = {}
        self._notifyHook = None
        self._reloader = self.getReloader()

    def getReloader(self):
        """Get the current reloader if the application is monitored."""
        reloader = None
        with warnings.catch_warnings():
            # ignore deprecation warnings from hupper
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            try:
                import hupper
            except ImportError:
                pass
            else:
                if hupper.is_active():
                    try:
                        reloader = hupper.get_reloader()
                    except RuntimeError:
                        pass
        if reloader:
            print('Application is monitored for reloading.')
            print()
        return reloader

    def moduleFromSpec(self, spec):
        """Load the module with the given module spec."""
        if not spec or not isinstance(spec, ModuleSpec):
            raise TypeError(f'Invalid spec: {spec!r}')
        try:
            module = module_from_spec(spec)
        except Exception:
            # Also record file paths which weren't successfully loaded, which
            # may happen due to a syntax error in a servlet, because we also
            # want to know when such a file is modified:
            if spec.origin:
                self.recordFile(spec.origin)
            raise
        self.recordModule(module)
        spec.loader.exec_module(module)
        return module

    def findSpec(self, name, path, fullModuleName=None):
        """Find the module spec for the given name at the given path."""
        if not name or not isinstance(name, str):
            raise TypeError(f'Invalid name: {name!r}')
        if not path or not isinstance(path, str):
            raise TypeError(f'Invalid path: {path!r}')

        # find package
        packageDirectory = join(path, name)
        packageFileName = '__init__.py'
        filePath = join(packageDirectory, packageFileName)
        if isfile(filePath):
            return spec_from_file_location(name, filePath)

        # find package as byte code without source
        filePath += 'c'
        if isfile(filePath):
            return spec_from_file_location(name, filePath)

        # find module
        fileName = f'{name}.py'
        filePath = join(path, fileName)
        if isfile(filePath):
            if fullModuleName:
                name = fullModuleName
            loader = SourceFileLoader(name, filePath)
            return spec_from_file_location(name, filePath, loader=loader)

        # find module as optimized byte code without source
        filePath += 'c'
        if isfile(filePath):
            loader = SourcelessFileLoader(name, filePath)
            return spec_from_file_location(name, filePath, loader=loader)

        raise ImportError(f'No module named {name!r}')

    def fileList(self, update=True):
        """Return the list of tracked files."""
        if update:
            # Update list of files of imported modules
            moduleNames = [modname for modname in sys.modules
                           if modname not in self._moduleFiles]
            if moduleNames:
                self.recordModules(moduleNames)
        return self._fileList

    def notifyOfNewFiles(self, hook):
        """Register notification hook.

        Called by someone else to register that they'd like to know
        when a new file is imported.
        """
        self._notifyHook = hook

    def watchFile(self, path, moduleName=None, getmtime=getmtime):
        """Add more files to watch without importing them."""
        mtime = getmtime(path)
        self._fileList[path] = mtime
        if moduleName:
            self._moduleFiles[moduleName] = path
        # send notification that this file was imported
        if self._notifyHook:
            self._notifyHook(path, mtime)
        # let reloader know that this was imported
        if self._reloader:
            self._reloader.watch_files([path])

    def recordModule(self, module, isfile=isfile):
        """Record a module."""
        moduleName = getattr(module, '__name__', None)
        if not moduleName or moduleName not in sys.modules:
            return
        fileList = self._fileList
        # __orig_file__ is used for PSP and Cheetah templates; we want
        # to record the source filenames, not the auto-generated modules:
        f = getattr(module, '__orig_file__', None)
        if f and f not in fileList:
            try:
                if isfile(f):
                    self.watchFile(f, moduleName)
            except OSError:
                pass
        else:
            f = getattr(module, '__file__', None)
            if f and f not in fileList:
                # record the .py file corresponding to each .pyc
                if f[-4:].lower() == '.pyc':
                    f = f[:-1]
                try:
                    if isfile(f):
                        self.watchFile(f, moduleName)
                    else:
                        self.watchFile(join(f, '__init__.py'))
                except OSError:
                    pass

    def recordModules(self, moduleNames=None):
        """Record a list of modules (or all modules)."""
        if moduleNames is None:
            moduleNames = sys.modules
        for modname in moduleNames:
            mod = sys.modules[modname]
            self.recordModule(mod)

    def recordFile(self, filename, isfile=isfile):
        """Record a file."""
        if isfile(filename):
            self.watchFile(filename)

    def fileUpdated(self, filename, update=True, getmtime=getmtime):
        """Check whether file has been updated."""
        fileList = self.fileList(update)
        try:
            oldTime = fileList[filename]
        except KeyError:
            return True
        try:
            newTime = getmtime(filename)
        except OSError:
            return True
        if oldTime >= newTime:
            return False
        fileList[filename] = newTime
        # Note that the file list could be changed while running this
        # method in a monitor thread, so we don't use iteritems() here:
        for moduleName, moduleFile in self._moduleFiles.items():
            if moduleFile == filename:
                module = sys.modules.get(moduleName)
                return not module or not getattr(
                    module, '__donotreload__', False)
        return True  # it's not a module, we must reload

    def updatedFile(self, update=True, getmtime=getmtime):
        """Check whether one of the files has been updated."""
        fileList = self.fileList(update)
        for filename, oldTime in list(fileList.items()):
            try:
                newTime = getmtime(filename)
            except OSError:
                return filename
            if oldTime >= newTime:
                continue
            fileList[filename] = newTime
            for moduleName, moduleFile in self._moduleFiles.items():
                if moduleFile == filename:
                    module = sys.modules.get(moduleName)
                    if module and getattr(module, '__donotreload__', False):
                        break
                    return filename  # it's a module that needs to be reloaded
            else:
                return filename  # it's not a module, we must reload

    def delModules(self, includePythonModules=False, excludePrefixes=None):
        """Delete imported modules.

        Deletes all the modules that have been imported unless they are part
        of Webware. This can be used to support auto reloading.
        """
        moduleFiles = self._moduleFiles
        fileList = self._fileList
        for modname in moduleFiles:
            if modname == __name__:
                continue
            filename = self._moduleFiles[modname]
            if not includePythonModules:
                if not filename or filename.startswith(sys.prefix):
                    continue
            for prefix in excludePrefixes or []:
                if modname.startswith(prefix):
                    break
            else:
                try:
                    del sys.modules[modname]
                except KeyError:
                    pass
                try:
                    del moduleFiles[modname]
                except KeyError:
                    pass
                try:
                    del fileList[filename]
                except KeyError:
                    pass



"""AJAX page template class

Written by John Dickinson based on ideas from
Apple Developer Connection and DivMod Nevow.
Some changes by Robert Forkel and Christoph Zwerschke.
"""

from time import time
from traceback import print_exc
from random import randrange

from io import StringIO

from .ExamplePage import ExamplePage as BaseClass


def quoteJs(what):
    """Return quoted JavaScript string corresponding to the Python object."""
    if isinstance(what, bool):
        ret = str(what).lower()
    elif isinstance(what, (int, float, PyJs)):
        ret = str(what)
    else:
        ret = "'{}'".format(str(what).replace('\\', '\\\\').replace(
            '\'', '\\\'').replace('\n', '\\n'))
    return ret


class PyJs:
    """Translate Python expressions to JavaScript strings."""

    def __init__(self, name):
        self._name = name

    def __getattr__(self, attr):
        return self.__class__(f'{self}.{attr}')

    def __str__(self):
        return self._name

    def __call__(self, *args, **kw):
        args = ','.join(map(quoteJs, args))
        kwArgs = ','.join(f'{k}={quoteJs(v)}' for k, v in kw.items())
        if args and kwArgs:
            allArgs = f'{args},{kwArgs}'
        elif not kwArgs:
            allArgs = args
        elif not args:
            allArgs = kwArgs
        return self.__class__(f'{self}({allArgs}')

    def __getitem__(self, index):
        return self.__class__(f'{self}[{quoteJs(index)}]')

    def __repr__(self):
        return self.__str__()


class AjaxPage(BaseClass):
    """A superclass for Webware servlets using Ajax techniques.

    AjaxPage can be used to make coding XMLHttpRequest() applications easier.

    Subclasses should override the method exposedMethods() which returns a list
    of method names. These method names refer to Webware Servlet methods that
    are able to be called by an Ajax-enabled web page. This is very similar
    in functionality to Webware's actions.

    A polling mechanism can be used for long running requests (e.g. generating
    reports) or if you want to send commands to the client without the client
    first triggering an event (e.g. for a chat application). In the first case,
    you should also specify a timeout after which polling shall be used.
    """

    # Class level variables that can be overridden by servlet instances:
    _debug = False  # set to True if you want to see debugging output
    _clientPolling = True  # set to True if you want to use polling
    _responseTimeout = 90  # timeout of client waiting for response in seconds

    # Class level variables to help make client code simpler:
    window, document, alert, this = map(
        PyJs, 'window document alert this'.split())
    setTag, setClass, setID, setValue, setReadonly = map(
        PyJs, 'setTag setClass setID setValue setReadonly'.split())
    call, callForm = map(PyJs, ('ajax_call', 'ajax_call_form'))

    # Response Queue for timed out queries:
    _responseQueue = {}

    def writeJavaScript(self):
        BaseClass.writeJavaScript(self)
        s = '<script src="ajax{}.js"></script>'
        self.writeln(s.format('call'))
        if self._clientPolling:
            self.writeln(s.format('poll'))

    def actions(self):
        actions = BaseClass.actions(self)
        actions.append('ajaxCall')
        if self._clientPolling:
            actions.append('ajaxPoll')
        return actions

    def exposedMethods(self):
        return []

    @staticmethod
    def clientPollingInterval():
        """Set the interval for the client polling.

        You should always make it a little random to avoid synchronization.
        """
        return randrange(3, 8)

    def ajaxCall(self):
        """Execute method with arguments on the server side.

        The method name is passed in the field _call_,
        the unique request number in the field _req_
        and the arguments in the field _ (single underscore).

        Returns JavaScript function to be executed by the client immediately.
        """
        req = self.request()
        startTime = None
        if req.hasField('_call_'):
            call = req.field('_call_')
            args = req.field('_', [])
            if not isinstance(args, list):
                args = [args]
            if self._clientPolling and self._responseTimeout:
                startTime = time()
            if call in self.exposedMethods():
                try:
                    method = getattr(self, call)
                except AttributeError:
                    cmd = self.alert(
                        f'{call}, although an approved method, was not found')
                else:
                    try:
                        if self._debug:
                            self.log(f"Ajax call {call}({args})")
                        cmd = str(method(*args))
                    except Exception:
                        err = StringIO()
                        print_exc(file=err)
                        e = err.getvalue()
                        cmd = self.alert(
                            f'{call} was called,'
                            f' but encountered an error: {e}')
                        err.close()
            else:
                cmd = self.alert(f'{call} is not an approved method')
        else:
            cmd = self.alert('Ajax call missing call parameter.')
        inTime = startTime is None or (
            time() - startTime < self._responseTimeout)
        if inTime:
            # If the computation of the method did not last very long,
            # deliver it immediately back to the client with this response:
            if self._debug:
                self.log(f"Ajax returns immediately: {cmd}")
            self.response().setHeader('Content-Type', 'application/json')
            self.write(cmd)
        else:
            # If the client request might have already timed out,
            # put the result in the queue and let client poll it:
            if self._debug:
                self.log(f"Ajax puts in queue: {cmd}")
            sid = self.session().identifier()
            self._responseQueue.setdefault(sid, []).append(cmd)

    def ajaxPoll(self):
        """Return queued JavaScript functions to be executed on the client.

        This is polled by the client in random intervals in order to get
        results from long-running queries or push content to the client.
        """
        if self._clientPolling:
            sid = self.session().identifier()
            # Set the timeout until the next time this method is called
            # by the client, using the Javascript wait variable:
            cmd = [f'wait={self.clientPollingInterval()}']
            if sid in self._responseQueue:  # add in other commands
                cmd.extend(map(str, self._responseQueue[sid]))
                self._responseQueue[sid] = []
            cmd = ';'.join(cmd) + ';'
            if self._debug:
                self.log(f"Ajax returns from queue: {cmd}")
        else:
            if self._debug:
                self.log("Ajax tells the client to stop polling.")
            cmd = 'dying=true;'
        self.response().setHeader('Content-Type', 'application/json')
        self.write(cmd)  # write out at least the wait variable

    def ajaxPush(self, cmd):
        """Push JavaScript commands to be executed on the client.

        Client polling must be activated if you want to use this.
        """
        if self._clientPolling:
            if self._debug:
                self.log(f"Ajax pushes in queue: {cmd}")
            sid = self.session().identifier()
            self._responseQueue.setdefault(sid, []).append(cmd)

    def preAction(self, actionName):
        if actionName not in ('ajaxCall', 'ajaxPoll'):
            BaseClass.preAction(self, actionName)

    def postAction(self, actionName):
        if actionName not in ('ajaxCall', 'ajaxPoll'):
            BaseClass.postAction(self, actionName)



from MiscUtils.Funcs import uniqueId

from .AdminPage import AdminPage


class LoginPage(AdminPage):
    """The log-in screen for the admin pages."""

    def writeContent(self):
        if self.loginDisabled():
            self.write(self.loginDisabled())
            return
        self.writeln(
            '<div style="margin-left:auto;margin-right:auto;width:20em">'
            '<p>&nbsp;</p>')
        extra = self.request().field('extra', None)
        if extra:
            self.writeln(f'<p style="color:#339">{self.htmlEncode(extra)}</p>')
        self.writeln('''<p>Please log in to view Administration Pages.
The username is <code>admin</code>. The password can be set in the
<code>Application.config</code> file in the <code>Configs</code> directory.</p>
<form method="post">
<table style="background-color:#cce;border:1px solid #33c;width:20em">
<tr><td style="text-align:right"><label for="username">Username:</label></td>
<td><input type="text" id="username" name="username" value="admin"></td></tr>
<tr><td style="text-align:right"><label for="password">Password:</label></td>
<td><input type="password" id="password" name="password" value=""></td></tr>
<tr><td colspan="2" style="text-align:right">
<input type="submit" name="login" value="Login"></td></tr>
</table>''')
        enc = self.htmlEncode
        for name, value in self.request().fields().items():
            if name not in ('login', 'logout', 'loginId',
                            'username', 'password', 'extra'):
                if not isinstance(value, list):
                    value = [value]
                for v in value:
                    self.writeln(f'<input type="hidden"'
                                 f' name="{enc(name)}" value="{enc(v)}">')
        if self.session().hasValue('loginId'):
            loginId = self.session().value('loginId')
        else:
            # Create a "unique" login id and put it in the form as well as
            # in the session. Login will only be allowed if they match.
            loginId = uniqueId(self)
            self.session().setValue('loginId', loginId)
        self.writeln(f'<input type="hidden" name="loginId" value="{loginId}">')
        self.writeln('</form>\n<p>&nbsp;</p></div>')



from os.path import exists, splitext

from .AdminSecurity import AdminSecurity


class View(AdminSecurity):
    """View a text or html file.

    The Admin View servlet loads any text or html file
    in your application working directory on the Webware server
    and displays it in the browser for your viewing pleasure.
    """

    def defaultAction(self):
        self._data = self._type = None
        self.writeHTML()
        if self._data and self._type:
            try:
                response = self.response()
                response.reset()
                response.setHeader('Content-Type', self._type)
                self.write(self._data)
            except Exception:
                self.writeError('File cannot be viewed!')

    def writeError(self, message):
        self.writeln(
            f'<h3 style="color:red">Error</h3><p>{message}</p>')

    def writeContent(self):
        filename = self.request().field('filename', None)
        if not filename:
            self.writeError('No filename given to view!')
            return
        filename = self.application().serverSidePath(filename)
        if not exists(filename):
            self.writeError(
                f'The requested file {filename!r} does not exist'
                ' in the server side directory.')
            return
        self._type = 'text/' + (
            'html' if splitext(filename)[1] in ('.htm', '.html') else 'plain')
        try:
            with open(filename) as f:
                self._data = f.read()
        except Exception:
            self.writeError(f'The requested file {filename!r} cannot be read.')
